from aana.exceptions.core import BaseException

__all__ = [
    "InferenceException",
    "MultipleFileUploadNotAllowed",
    "PromptTooLongException",
    "EndpointNotFoundException",
    "TooManyRequestsException",
    "HandlerAlreadyRegisteredException",
    "HandlerNotRegisteredException",
]


class InferenceException(BaseException):
    """Exception raised when there is an error during inference.

    Attributes:
        model_name -- name of the model
    """

    def __init__(self, model_name):
        """Initialize the exception.

        Args:
            model_name (str): name of the model that caused the exception
        """
        super().__init__(model_name=model_name)
        self.model_name = model_name

    def __reduce__(self):
        """Called when the exception is pickled.

        We need to do this if exception has one or more arguments.
        See https://bugs.python.org/issue32696#msg310963 for more info.
        """
        # TODO: check if there is a better way to do this
        return (self.__class__, (self.model_name,))


class MultipleFileUploadNotAllowed(BaseException):
    """Exception raised when multiple inputs require file upload.

    Attributes:
        input_name -- name of the input
    """

    def __init__(self, input_name: str):
        """Initialize the exception.

        Args:
            input_name (str): name of the input that caused the exception
        """
        super().__init__(input_name=input_name)
        self.input_name = input_name

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.input_name,))


class PromptTooLongException(BaseException):
    """Exception raised when the prompt is too long.

    Attributes:
        prompt_len (int): the length of the prompt in tokens
        max_len (int): the maximum allowed length of the prompt in tokens
    """

    def __init__(self, prompt_len: int, max_len: int):
        """Initialize the exception.

        Args:
            prompt_len (int): the length of the prompt in tokens
            max_len (int): the maximum allowed length of the prompt in tokens
        """
        super().__init__(prompt_len=prompt_len, max_len=max_len)
        self.prompt_len = prompt_len
        self.max_len = max_len

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.prompt_len, self.max_len))


class EndpointNotFoundException(BaseException):
    """Exception raised when an endpoint is not found.

    Attributes:
        target (str): the name of the target deployment
        endpoint (str): the endpoint path
    """

    def __init__(self, target: str, endpoint: str):
        """Initialize the exception.

        Args:
            target (str): the name of the target deployment
            endpoint (str): the endpoint path
        """
        super().__init__(target=target, endpoint=endpoint)
        self.target = target
        self.endpoint = endpoint

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.target, self.endpoint))


class TooManyRequestsException(BaseException):
    """Exception raised when calling a rate-limited resource too often.

    Attributes:
        rate_limit (int): The limit amount.
        rate_duration (float): The duration for the limit in seconds.
    """

    def __init__(self, rate_limit: int, rate_duration: float):
        """Constructor.

        Args:
            rate_limit (int): The limit amount.
            rate_duration (float): The duration for the limit in seconds.
        """
        super().__init__(rate_limit=rate_limit, rate_duration=rate_duration)
        self.rate_limit = rate_limit
        self.rate_duration = rate_duration

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.rate_limit, self.rate_duration))


class HandlerAlreadyRegisteredException(BaseException):
    """Exception raised when registering a handler that is already registered."""

    def __init__(self):
        """Constructor."""
        super().__init__()

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, ())


class HandlerNotRegisteredException(BaseException):
    """Exception removing a handler that has not been registered."""

    def __init__(self):
        """Constructor."""
        super().__init__()

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, ())


class EmptyMigrationsException(BaseException):
    """Exception raised when there are no migrations to apply."""

    pass


class DeploymentException(Exception):
    """Base exception for deployment errors."""

    pass


class InsufficientResources(DeploymentException):
    """Exception raised when there are insufficient resources for a deployment."""

    pass


class FailedDeployment(DeploymentException):
    """Exception raised when there is an error during deployment."""

    pass



# ruff: noqa: S101
from importlib import resources

import pytest
from hqq.core.quantize import BaseQuantizeConfig

from aana.core.models.sampling import SamplingParams
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.hqq_text_generation_deployment import (
    HqqBackend,
    HqqTexGenerationConfig,
    HqqTextGenerationDeployment,
)
from aana.tests.utils import verify_deployment_results
from aana.utils.core import get_object_hash

deployments = [
    (
        (
            "meta-llama/Meta-Llama-3.1-8B-Instruct",
            HqqTextGenerationDeployment.options(
                num_replicas=1,
                ray_actor_options={"num_gpus": 0.5},
                user_config=HqqTexGenerationConfig(
                    model_id="meta-llama/Meta-Llama-3.1-8B-Instruct",
                    backend=HqqBackend.TORCHAO_INT4,
                    compile=True,
                    quantize_on_fly=True,
                    quantization_config=BaseQuantizeConfig(
                        nbits=4, group_size=64, axis=1
                    ),
                    default_sampling_params=SamplingParams(
                        temperature=0.0, top_p=1.0, top_k=-1, max_tokens=512
                    ),
                    model_kwargs={"attn_implementation": "sdpa"},
                ).model_dump(mode="json"),
            ),
        ),
        "<s><|user|>\n{query}<|end|>\n<|assistant|>\n",
    ),
    (
        (
            "mobiuslabsgmbh/Llama-3.1-8b-instruct_4bitgs64_hqq_calib",
            HqqTextGenerationDeployment.options(
                num_replicas=1,
                max_ongoing_requests=1000,
                ray_actor_options={"num_gpus": 0.5},
                user_config=HqqTexGenerationConfig(
                    model_id="mobiuslabsgmbh/Llama-3.1-8b-instruct_4bitgs64_hqq_calib",
                    backend=HqqBackend.TORCHAO_INT4,
                    compile=False,
                    quantize_on_fly=False,
                    quantization_config=BaseQuantizeConfig(
                        nbits=4,
                        group_size=64,
                        quant_scale=False,
                        quant_zero=False,
                        axis=1,
                    ),
                    default_sampling_params=SamplingParams(
                        temperature=0.0, top_p=1.0, top_k=-1, max_tokens=512
                    ),
                ).model_dump(mode="json"),
            ),
        ),
        "<s><|user|>\n{query}<|end|>\n<|assistant|>\n",
    ),
    (
        (
            "meta-llama/Meta-Llama-3.1-8B-Instruct",
            HqqTextGenerationDeployment.options(
                num_replicas=1,
                ray_actor_options={"num_gpus": 0.5},
                user_config=HqqTexGenerationConfig(
                    model_id="meta-llama/Meta-Llama-3.1-8B-Instruct",
                    backend=HqqBackend.BITBLAS,
                    compile=True,
                    quantize_on_fly=True,
                    quantization_config=BaseQuantizeConfig(
                        nbits=4, group_size=64, axis=1
                    ),
                    default_sampling_params=SamplingParams(
                        temperature=0.0, top_p=1.0, top_k=-1, max_tokens=512
                    ),
                    model_kwargs={"attn_implementation": "sdpa"},
                ).model_dump(mode="json"),
            ),
        ),
        "<s><|user|>\n{query}<|end|>\n<|assistant|>\n",
    ),
    (
        (
            "mobiuslabsgmbh/Llama-3.1-8b-instruct_4bitgs64_hqq_calib",
            HqqTextGenerationDeployment.options(
                num_replicas=1,
                max_ongoing_requests=1000,
                ray_actor_options={"num_gpus": 0.5},
                user_config=HqqTexGenerationConfig(
                    model_id="mobiuslabsgmbh/Llama-3.1-8b-instruct_4bitgs64_hqq_calib",
                    backend=HqqBackend.BITBLAS,
                    compile=False,
                    quantize_on_fly=False,
                    quantization_config=BaseQuantizeConfig(
                        nbits=4,
                        group_size=64,
                        quant_scale=False,
                        quant_zero=False,
                        axis=1,
                    ),
                    default_sampling_params=SamplingParams(
                        temperature=0.0, top_p=1.0, top_k=-1, max_tokens=512
                    ),
                ).model_dump(mode="json"),
            ),
        ),
        "<s><|user|>\n{query}<|end|>\n<|assistant|>\n",
    ),
]


@pytest.mark.parametrize(
    "setup_deployment, prompt_template", deployments, indirect=["setup_deployment"]
)
class TestHQQDeployments:
    """Test hqq deployments."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize("query", ["Who is Elon Musk?"])
    async def test_hqq_text_generation_methods(
        self, setup_deployment, prompt_template, query
    ):
        """Test text generation methods."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        query_hash = get_object_hash(query)
        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "hqq_generation"
            / f"{deployment_name}_{query_hash}.json"
        )

        prompt = prompt_template.format(query=query)

        # test generate method
        output = await handle.generate(
            prompt=prompt,
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )
        verify_deployment_results(expected_output_path, output["text"])

        # test generate_stream method
        stream = handle.generate_stream(
            prompt=prompt,
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )
        text = ""
        async for chunk in stream:
            text += chunk["text"]

        verify_deployment_results(expected_output_path, text)

        # test generate_batch method
        output = await handle.generate_batch(
            prompts=[prompt, prompt],
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )
        texts = output["texts"]

        for text in texts:
            verify_deployment_results(expected_output_path, text)



# ruff: noqa: S101, S113

import pytest
from ray import serve

from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.base_deployment import BaseDeployment, exception_handler


@serve.deployment(health_check_period_s=1, health_check_timeout_s=30)
class Lowercase(BaseDeployment):
    """Ray deployment that returns the lowercase version of a text."""

    def __init__(self):
        """Initialize the deployment."""
        super().__init__()
        self.num_requests = 0

    @exception_handler
    async def lower(self, text: str) -> dict:
        """Lowercase the text.

        Args:
            text (str): The text to lowercase

        Returns:
            dict: The lowercase text
        """
        # Only every 3rd request should be successful
        self.num_requests += 1
        if self.num_requests % 3 != 0:
            raise Exception("Random exception")  # noqa: TRY002, TRY003

        return {"text": text.lower()}


deployments = [
    {
        "name": "lowercase_deployment",
        "instance": Lowercase,
    }
]


@pytest.mark.asyncio
async def test_deployment_retry(create_app):
    """Test the Ray Serve app."""
    create_app(deployments, [])

    text = "Hello, World!"

    # Get deployment handle without retries
    handle = await AanaDeploymentHandle.create(
        "lowercase_deployment", retry_exceptions=False
    )

    # test the lowercase deployment fails
    with pytest.raises(Exception):  # noqa: B017
        await handle.lower(text=text)

    # Get deployment handle with retries
    handle = await AanaDeploymentHandle.create(
        "lowercase_deployment", retry_exceptions=True
    )

    # test the lowercase deployment works
    response = await handle.lower(text=text)
    assert response == {"text": text.lower()}



# ruff: noqa: S101

import asyncio
from datetime import datetime, timedelta, timezone

import pytest

from aana.storage.models.task import Status as TaskStatus
from aana.storage.models.task import TaskEntity
from aana.storage.repository.task import TaskRepository


def test_save_task_repo(db_session):
    """Test saving a task."""
    task_repo = TaskRepository(db_session)
    task = task_repo.save(endpoint="/test", data={"test": "test"})

    task_entity = task_repo.read(task.id)
    assert task_entity
    assert task_entity.id == task.id
    assert task_entity.endpoint == "/test"
    assert task_entity.data == {"test": "test"}


@pytest.mark.asyncio
async def test_multiple_simultaneous_tasks(db_session):
    """Test creating multiple tasks in parallel."""
    task_repo = TaskRepository(db_session)

    # Create multiple tasks in parallel with asyncio
    async def add_task(i):
        task = task_repo.save(endpoint="/test", data={"test": i})
        return task

    async_tasks = []
    for i in range(30):
        async_task = asyncio.create_task(add_task(i))
        async_tasks.append(async_task)

    # Wait for all tasks to complete
    await asyncio.gather(*async_tasks)


def test_get_unprocessed_tasks(db_session):
    """Test fetching unprocessed tasks."""
    task_repo = TaskRepository(db_session)

    def _create_sample_tasks():
        # Remove all existing tasks
        db_session.query(TaskEntity).delete()
        db_session.commit()

        # Create sample tasks with different statuses
        now = datetime.now(timezone.utc)

        task1 = TaskEntity(
            endpoint="/test1",
            data={"test": "data1"},
            status=TaskStatus.CREATED,
            priority=1,
            created_at=now - timedelta(hours=10),
        )
        task2 = TaskEntity(
            endpoint="/test2",
            data={"test": "data2"},
            status=TaskStatus.NOT_FINISHED,
            priority=2,
            created_at=now - timedelta(hours=1),
        )
        task3 = TaskEntity(
            endpoint="/test3",
            data={"test": "data3"},
            status=TaskStatus.COMPLETED,
            priority=3,
            created_at=now - timedelta(hours=2),
        )
        task4 = TaskEntity(
            endpoint="/test4",
            data={"test": "data4"},
            status=TaskStatus.CREATED,
            priority=2,
            created_at=now - timedelta(hours=3),
        )
        task5 = TaskEntity(
            endpoint="/test5",
            data={"test": "data5"},
            status=TaskStatus.FAILED,
            priority=1,
            created_at=now - timedelta(minutes=1),
            result={"error": "InferenceException"},
        )
        task6 = TaskEntity(
            endpoint="/test6",
            data={"test": "data6"},
            status=TaskStatus.RUNNING,
            priority=3,
            created_at=now - timedelta(minutes=2),
        )
        task7 = TaskEntity(
            endpoint="/test7",
            data={"test": "data7"},
            status=TaskStatus.FAILED,
            priority=1,
            created_at=now - timedelta(minutes=3),
            result={"error": "NonRecoverableError"},
        )

        db_session.add_all([task1, task2, task3, task4, task5, task6, task7])
        db_session.commit()
        return task1, task2, task3, task4, task5, task6, task7

    # Create sample tasks
    task1, task2, task3, task4, task5, task6, task7 = _create_sample_tasks()

    # Fetch unprocessed tasks without any limit
    unprocessed_tasks = task_repo.fetch_unprocessed_tasks(
        max_retries=3, retryable_exceptions=["InferenceException"]
    )

    # Assert that only tasks with CREATED and NOT_FINISHED status are returned
    assert len(unprocessed_tasks) == 4
    assert task1 in unprocessed_tasks
    assert task2 in unprocessed_tasks
    assert task4 in unprocessed_tasks
    assert task5 in unprocessed_tasks

    # Ensure tasks are ordered by priority and then by created_at
    assert unprocessed_tasks[0].id == task4.id  # Highest priority
    assert unprocessed_tasks[1].id == task2.id  # Same priority, but a newer task
    assert unprocessed_tasks[2].id == task1.id  # Lowest priority
    assert unprocessed_tasks[3].id == task5.id  # Highest priority, but older

    # Create sample tasks
    task1, task2, task3, task4, task5, task6, task7 = _create_sample_tasks()

    # Fetch unprocessed tasks with a limit
    limited_tasks = task_repo.fetch_unprocessed_tasks(
        limit=2, max_retries=3, retryable_exceptions=["InferenceException"]
    )

    # Assert that only the specified number of tasks is returned
    assert len(limited_tasks) == 2
    assert limited_tasks[0].id == task4.id  # Highest priority
    assert limited_tasks[1].id == task2.id  # Same priority, but older


def test_update_status(db_session):
    """Test updating the status of a task."""
    task_repo = TaskRepository(db_session)

    # Create a task with an initial status
    task = TaskEntity(
        endpoint="/test", data={"key": "value"}, status=TaskStatus.CREATED
    )
    db_session.add(task)
    db_session.commit()

    # Update the status to ASSIGNED and check fields
    task_repo.update_status(task.id, TaskStatus.ASSIGNED, progress=50)

    updated_task = task_repo.read(task.id)
    assert updated_task.status == TaskStatus.ASSIGNED
    assert updated_task.assigned_at is not None
    assert updated_task.num_retries == 1
    assert updated_task.progress == 50
    assert updated_task.result is None

    # Update the status to COMPLETED and check fields
    task_repo.update_status(
        task.id, TaskStatus.COMPLETED, progress=100, result={"result": "final_result"}
    )

    updated_task = task_repo.read(task.id)
    assert updated_task.status == TaskStatus.COMPLETED
    assert updated_task.completed_at is not None
    assert updated_task.progress == 100
    assert updated_task.result == {"result": "final_result"}

    # Ensure timestamps are reasonable
    assert updated_task.assigned_at < updated_task.completed_at
    assert updated_task.created_at < updated_task.assigned_at

    # Update the status to FAILED and check fields
    task_repo.update_status(
        task.id, TaskStatus.FAILED, progress=0, result={"error": "error_message"}
    )
    updated_task = task_repo.read(task.id)

    assert updated_task.status == TaskStatus.FAILED
    assert updated_task.completed_at is not None
    assert updated_task.progress == 0
    assert updated_task.result == {"error": "error_message"}

    # Ensure timestamps are reasonable
    assert updated_task.assigned_at < updated_task.completed_at
    assert updated_task.created_at < updated_task.assigned_at


def test_get_active_tasks(db_session):
    """Test fetching active tasks."""
    task_repo = TaskRepository(db_session)

    # Remove all existing tasks
    db_session.query(TaskEntity).delete()
    db_session.commit()

    # Create sample tasks with different statuses
    task1 = TaskEntity(
        endpoint="/task1", data={"test": "data1"}, status=TaskStatus.CREATED
    )
    task2 = TaskEntity(
        endpoint="/task2", data={"test": "data2"}, status=TaskStatus.RUNNING
    )
    task3 = TaskEntity(
        endpoint="/task3", data={"test": "data3"}, status=TaskStatus.ASSIGNED
    )
    task4 = TaskEntity(
        endpoint="/task4", data={"test": "data4"}, status=TaskStatus.COMPLETED
    )
    task5 = TaskEntity(
        endpoint="/task5", data={"test": "data5"}, status=TaskStatus.FAILED
    )
    task6 = TaskEntity(
        endpoint="/task6", data={"test": "data6"}, status=TaskStatus.NOT_FINISHED
    )

    db_session.add_all([task1, task2, task3, task4, task5, task6])
    db_session.commit()

    # Fetch active tasks
    active_tasks = task_repo.get_active_tasks()

    # Assert that only tasks with RUNNING and ASSIGNED status are returned
    assert len(active_tasks) == 2
    assert task2 in active_tasks
    assert task3 in active_tasks
    assert all(
        task.status in [TaskStatus.RUNNING, TaskStatus.ASSIGNED]
        for task in active_tasks
    )


def test_remove_completed_tasks(db_session):
    """Test removing completed tasks."""
    task_repo = TaskRepository(db_session)

    # Remove all existing tasks
    db_session.query(TaskEntity).delete()
    db_session.commit()

    # Create sample tasks with different statuses
    task1 = TaskEntity(
        endpoint="/task1", data={"test": "data1"}, status=TaskStatus.COMPLETED
    )
    task2 = TaskEntity(
        endpoint="/task2", data={"test": "data2"}, status=TaskStatus.RUNNING
    )
    task3 = TaskEntity(
        endpoint="/task3", data={"test": "data3"}, status=TaskStatus.ASSIGNED
    )
    task4 = TaskEntity(
        endpoint="/task4", data={"test": "data4"}, status=TaskStatus.COMPLETED
    )
    task5 = TaskEntity(
        endpoint="/task5", data={"test": "data5"}, status=TaskStatus.FAILED
    )
    task6 = TaskEntity(
        endpoint="/task6", data={"test": "data6"}, status=TaskStatus.NOT_FINISHED
    )

    all_tasks = [task1, task2, task3, task4, task5, task6]
    unfinished_tasks = [task2, task3]

    db_session.add_all(all_tasks)
    db_session.commit()

    # Remove completed tasks
    task_ids = [str(task.id) for task in all_tasks]
    non_completed_task_ids = task_repo.filter_incomplete_tasks(task_ids)

    # Assert that only the task IDs that are not completed are returned
    assert set(non_completed_task_ids) == {str(task.id) for task in unfinished_tasks}


def test_update_expired_tasks(db_session):
    """Test updating expired tasks."""
    task_repo = TaskRepository(db_session)

    # Remove all existing tasks
    db_session.query(TaskEntity).delete()
    db_session.commit()

    # Set up current time and a cutoff time
    current_time = datetime.now(timezone.utc)
    execution_timeout = 3600  # 1 hour in seconds
    heartbeat_timeout = 60  # 1 minute in seconds

    # Create tasks with different updated_at times and statuses
    task1 = TaskEntity(
        endpoint="/task1",
        data={"test": "data1"},
        status=TaskStatus.RUNNING,
        assigned_at=current_time - timedelta(hours=2),
        updated_at=current_time - timedelta(seconds=10),
    )
    task2 = TaskEntity(
        endpoint="/task2",
        data={"test": "data2"},
        status=TaskStatus.ASSIGNED,
        assigned_at=current_time - timedelta(seconds=2),
        updated_at=current_time - timedelta(seconds=5),
    )
    task3 = TaskEntity(
        endpoint="/task3",
        data={"test": "data3"},
        status=TaskStatus.RUNNING,
        assigned_at=current_time - timedelta(seconds=2),
        updated_at=current_time,
    )
    task4 = TaskEntity(
        endpoint="/task4",
        data={"test": "data4"},
        status=TaskStatus.COMPLETED,
        assigned_at=current_time - timedelta(hours=1),
        updated_at=current_time - timedelta(hours=2),
    )
    task5 = TaskEntity(
        endpoint="/task5",
        data={"test": "data5"},
        status=TaskStatus.FAILED,
        assigned_at=current_time - timedelta(minutes=1),
        updated_at=current_time - timedelta(seconds=4),
    )
    task6 = TaskEntity(
        endpoint="/task6",
        data={"test": "data6"},
        status=TaskStatus.RUNNING,
        assigned_at=current_time - timedelta(minutes=3),
        updated_at=current_time - timedelta(minutes=2),
    )

    db_session.add_all([task1, task2, task3, task4, task5, task6])
    db_session.commit()

    # Fetch expired tasks
    expired_tasks = task_repo.update_expired_tasks(
        execution_timeout=execution_timeout,
        heartbeat_timeout=heartbeat_timeout,
        max_retries=3,
    )

    # Assert that only tasks with RUNNING or ASSIGNED status and an assigned_at time older than the execution_timeout or
    # heartbeat_timeout are returned
    expected_task_ids = {str(task1.id), str(task6.id)}
    returned_task_ids = {str(task.id) for task in expired_tasks}

    assert returned_task_ids == expected_task_ids



import datetime
from typing import Annotated, Any, TypeVar

from sqlalchemy import DateTime, MetaData, String, func
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    mapped_column,
    object_mapper,
    registry,
)

from aana.core.models.media import MediaId

timestamp = Annotated[
    datetime.datetime,
    mapped_column(DateTime(timezone=True)),
]

T = TypeVar("T", bound="InheritanceReuseMixin")


class InheritanceReuseMixin:
    """Mixin for instantiating child classes from parent instances."""

    @classmethod
    def from_parent(cls: type[T], parent_instance: Any, **kwargs: Any) -> T:
        """Create a new instance of the child class, reusing attributes from the parent instance.

        Args:
            parent_instance (Any): An instance of the parent class
            kwargs (Any): Additional keyword arguments to set on the new instance

        Returns:
            T: A new instance of the child class
        """
        # Get the mapped attributes of the parent class
        mapper = object_mapper(parent_instance)
        attributes = {
            prop.key: getattr(parent_instance, prop.key)
            for prop in mapper.iterate_properties
            if hasattr(parent_instance, prop.key)
            and prop.key
            != mapper.polymorphic_on.name  # don't copy the polymorphic_on attribute from the parent
        }

        # Update attributes with any additional kwargs
        attributes.update(kwargs)

        # Create and return a new instance of the child class
        return cls(**attributes)


class BaseEntity(DeclarativeBase, InheritanceReuseMixin):
    """Base for all ORM classes."""

    metadata = MetaData(
        naming_convention={
            "ix": "ix_%(column_0_label)s",
            "uq": "uq_%(table_name)s_%(column_0_name)s",
            "ck": "ck_%(table_name)s_`%(constraint_name)s`",
            "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
            "pk": "pk_%(table_name)s",
        }
    )

    registry = registry(
        type_annotation_map={
            MediaId: String(36),
        }
    )

    def __repr__(self) -> str:
        """Get the representation of the entity."""
        return f"{self.__class__.__name__}(id={self.id})"


class TimeStampEntity:
    """Mixin for database entities that will have create/update timestamps."""

    created_at: Mapped[timestamp] = mapped_column(
        server_default=func.now(),
        comment="Timestamp when row is inserted",
    )
    updated_at: Mapped[timestamp] = mapped_column(
        onupdate=func.now(),
        server_default=func.now(),
        comment="Timestamp when row is updated",
    )



# ruff: noqa: S101
from importlib import resources

import pytest

from aana.core.models.chat import ChatDialog, ChatMessage
from aana.core.models.sampling import SamplingParams
from aana.core.models.types import Dtype
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.hf_text_generation_deployment import (
    HfTextGenerationConfig,
    HfTextGenerationDeployment,
)
from aana.deployments.vllm_deployment import VLLMConfig, VLLMDeployment
from aana.exceptions.runtime import PromptTooLongException
from aana.tests.utils import verify_deployment_results
from aana.utils.core import get_object_hash

deployments = [
    (
        (
            "phi3_mini_4k_instruct_hf_text_generation_deployment",
            HfTextGenerationDeployment.options(
                num_replicas=1,
                ray_actor_options={"num_gpus": 0.25},
                user_config=HfTextGenerationConfig(
                    model_id="microsoft/Phi-3-mini-4k-instruct",
                    model_kwargs={
                        "trust_remote_code": True,
                    },
                    default_sampling_params=SamplingParams(
                        temperature=0.0, kwargs={"diversity_penalty": 0.0}
                    ),
                ).model_dump(mode="json"),
            ),
        ),
        "<s><|user|>\n{query}<|end|>\n<|assistant|>\n",
    ),
    (
        (
            "phi3_mini_4k_instruct_vllm_deployment",
            VLLMDeployment.options(
                num_replicas=1,
                max_ongoing_requests=1000,
                ray_actor_options={"num_gpus": 0.25},
                user_config=VLLMConfig(
                    model="microsoft/Phi-3-mini-4k-instruct",
                    dtype=Dtype.FLOAT16,
                    gpu_memory_reserved=10000,
                    enforce_eager=True,
                    default_sampling_params=SamplingParams(
                        temperature=0.0,
                        top_p=1.0,
                        top_k=-1,
                        max_tokens=1024,
                        kwargs={"frequency_penalty": 0.0},
                    ),
                    engine_args={
                        "trust_remote_code": True,
                    },
                ).model_dump(mode="json"),
            ),
        ),
        "<s><|user|>\n{query}<|end|>\n<|assistant|>\n",
    ),
]


@pytest.mark.parametrize(
    "setup_deployment, prompt_template", deployments, indirect=["setup_deployment"]
)
class TestTextGenerationDeployments:
    """Test text generation deployments."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize("query", ["Who is Elon Musk?"])
    async def test_text_generation_methods(
        self, setup_deployment, prompt_template, query
    ):
        """Test text generation methods."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        query_hash = get_object_hash(query)
        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "text_generation"
            / f"{deployment_name}_{query_hash}.json"
        )

        prompt = prompt_template.format(query=query)

        # test generate method
        output = await handle.generate(
            prompt=prompt,
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )
        verify_deployment_results(expected_output_path, output["text"])

        # test generate_stream method
        stream = handle.generate_stream(
            prompt=prompt,
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )
        text = ""
        async for chunk in stream:
            text += chunk["text"]

        verify_deployment_results(expected_output_path, text)

        # test generate_batch method
        output = await handle.generate_batch(
            prompts=[prompt, prompt],
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )
        texts = output["texts"]

        for text in texts:
            verify_deployment_results(expected_output_path, text)

        # test chat method
        dialog = ChatDialog(
            messages=[
                ChatMessage(
                    role="user",
                    content=query,
                )
            ]
        )
        output = await handle.chat(
            dialog=dialog,
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )

        response_message = output["message"]
        assert response_message.role == "assistant"
        text = response_message.content

        verify_deployment_results(expected_output_path, text)

        # test chat_stream method
        stream = handle.chat_stream(
            dialog=dialog,
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )

        text = ""
        async for chunk in stream:
            text += chunk["text"]

        verify_deployment_results(expected_output_path, text)

        # test generate method with too long prompt
        with pytest.raises(PromptTooLongException):
            output = await handle.generate(
                prompt=prompt * 1000,
                sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
            )



from uuid import uuid4

from sqlalchemy.orm import Mapped, mapped_column

from aana.core.models.media import MediaId
from aana.storage.models.base import BaseEntity, TimeStampEntity


class MediaEntity(BaseEntity, TimeStampEntity):
    """Base ORM class for media (e.g. videos, images, etc.).

    This class is meant to be subclassed by other media types.

    Attributes:
        id (MediaId): Unique identifier for the media.
        media_type (str): The type of media (populated automatically by ORM based on `polymorphic_identity` of   subclass).
    """

    __tablename__ = "media"
    id: Mapped[MediaId] = mapped_column(
        primary_key=True,
        default=lambda: str(uuid4()),
        comment="Unique identifier for the media",
    )
    media_type: Mapped[str] = mapped_column(comment="The type of media")

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "media",
        "polymorphic_on": "media_type",
    }



# ruff: noqa: S101
import pytest

from aana.core.models.whisper import WhisperParams


def test_whisper_params_default():
    """Test the default values of WhisperParams object.

    Keeping the default parameters of a function or object is important
    in case other code relies on them.

    If you need to change the default parameters, think twice before doing so.
    """
    params = WhisperParams()

    assert params.language is None
    assert params.beam_size == 5
    assert params.best_of == 5
    assert params.temperature == (0.0, 0.2, 0.4, 0.6, 0.8, 1.0)
    assert params.word_timestamps is False
    assert params.vad_filter is True


@pytest.mark.parametrize(
    "language, beam_size, best_of, temperature, word_timestamps, vad_filter",
    [
        ("en", 5, 5, 0.5, True, True),
        ("fr", 3, 3, 0.2, False, False),
        (None, 1, 1, [0.8, 0.9], True, False),
    ],
)
def test_whisper_params(
    language, beam_size, best_of, temperature, word_timestamps, vad_filter
):
    """Test function for the WhisperParams class with valid parameters."""
    params = WhisperParams(
        language=language,
        beam_size=beam_size,
        best_of=best_of,
        temperature=temperature,
        word_timestamps=word_timestamps,
        vad_filter=vad_filter,
    )

    assert params.language == language
    assert params.beam_size == beam_size
    assert params.best_of == best_of
    assert params.temperature == temperature
    assert params.word_timestamps == word_timestamps
    assert params.vad_filter == vad_filter


@pytest.mark.parametrize(
    "temperature",
    [
        [-1.0, 0.5, 1.5],
        [0.0, 0.2, 0.4, 0.6, 0.8, 1.0, 2.0],
        "invalid_temperature",
        2,
    ],
)
def test_whisper_params_invalid_temperature(temperature):
    """Check ValueError raised if temperature is invalid."""
    with pytest.raises(ValueError):
        WhisperParams(temperature=temperature)



# ruff: noqa: S101, NPY002
from importlib import resources
from pathlib import Path

import numpy as np
import PIL.Image
import pytest

from aana.core.models.image import Image


def load_numpy_from_image_bytes(content: bytes) -> np.ndarray:
    """Load a numpy array from image bytes."""
    image = Image(content=content, save_on_disk=False)
    return image.get_numpy()


@pytest.fixture
def mock_download_file(mocker):
    """Mock download_file function in both media and image modules."""
    # Path to the file to be used as mock return value
    path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
    content = path.read_bytes()

    # Mock for aana.core.models.media.download_file
    mock_media_download = mocker.patch(
        "aana.core.models.media.download_file", autospec=True
    )
    mock_media_download.return_value = content

    # Mock for aana.core.models.image.download_file
    mock_image_download = mocker.patch(
        "aana.core.models.image.download_file", autospec=True
    )
    mock_image_download.return_value = content

    # Return both mocks in case you need them in your tests
    return mock_media_download, mock_image_download


def test_image(mock_download_file):
    """Test that the image can be created from path, url, content, or numpy."""
    try:
        path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
        image = Image(path=path, save_on_disk=False)
        assert image.path == path
        assert image.content is None
        assert image.numpy is None
        assert image.url is None

        numpy = image.get_numpy()
        assert numpy.shape == (720, 909, 3)

        content = image.get_content()
        numpy = load_numpy_from_image_bytes(content)
        assert numpy.shape == (720, 909, 3)
    finally:
        image.cleanup()

    try:
        url = "http://example.com/Starry_Night.jpeg"
        image = Image(url=url, save_on_disk=False)
        assert image.path is None
        assert image.content is None
        assert image.numpy is None
        assert image.url == url

        numpy = image.get_numpy()
        assert numpy.shape == (720, 909, 3)

        content = image.get_content()
        numpy = load_numpy_from_image_bytes(content)
        assert numpy.shape == (720, 909, 3)
    finally:
        image.cleanup()

    try:
        path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
        content = path.read_bytes()
        image = Image(content=content, save_on_disk=False)
        assert image.path is None
        assert image.content == content
        assert image.numpy is None
        assert image.url is None

        numpy = image.get_numpy()
        assert numpy.shape == (720, 909, 3)

        content = image.get_content()
        numpy = load_numpy_from_image_bytes(content)
        assert numpy.shape == (720, 909, 3)
    finally:
        image.cleanup()

    try:
        numpy = np.random.rand(100, 100, 3).astype(np.uint8)
        image = Image(numpy=numpy, save_on_disk=False)
        assert image.path is None
        assert image.content is None
        assert np.array_equal(image.numpy, numpy)
        assert image.url is None

        numpy = image.get_numpy()
        assert np.array_equal(numpy, numpy)

        content = image.get_content()
        numpy = load_numpy_from_image_bytes(content)
        assert np.array_equal(numpy, numpy)
    finally:
        image.cleanup()


def test_image_path_not_exist():
    """Test that the image can't be created from path if the path doesn't exist."""
    path = Path("path/to/image_that_does_not_exist.jpeg")
    with pytest.raises(FileNotFoundError):
        Image(path=path)


def test_save_image(mock_download_file):
    """Test that save_on_disk works."""
    try:
        path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
        image = Image(path=path, save_on_disk=True)
        assert image.path == path
        assert image.content is None
        assert image.numpy is None
        assert image.url is None
        assert image.path.exists()
    finally:
        image.cleanup()
    # if image is provided as a path, cleanup() should NOT delete the image
    assert image.path.exists()

    try:
        url = "http://example.com/Starry_Night.jpeg"
        image = Image(url=url, save_on_disk=True)
        assert image.content is None
        assert image.numpy is None
        assert image.url == url
        assert image.path.exists()
    finally:
        image.cleanup()

    try:
        path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
        content = path.read_bytes()
        image = Image(content=content, save_on_disk=True)
        assert image.content == content
        assert image.numpy is None
        assert image.url is None
        assert image.path.exists()
    finally:
        image.cleanup()

    try:
        numpy = np.random.rand(100, 100, 3).astype(np.uint8)
        image = Image(numpy=numpy, save_on_disk=True)
        assert image.content is None
        assert np.array_equal(image.numpy, numpy)
        assert image.url is None
        assert image.path.exists()
    finally:
        image.cleanup()


def test_cleanup(mock_download_file):
    """Test that cleanup works."""
    try:
        url = "http://example.com/Starry_Night.jpeg"
        image = Image(url=url, save_on_disk=True)
        assert image.path.exists()
    finally:
        image.cleanup()
        assert not image.path.exists()


def test_at_least_one_input():
    """Test that at least one input is provided."""
    with pytest.raises(ValueError):
        Image(save_on_disk=False)

    with pytest.raises(ValueError):
        Image(save_on_disk=True)


def test_get_pil_image():
    """Test that get_pil_image method."""
    try:
        path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
        image = Image(path=path, save_on_disk=False)
        pil_image = image.get_pil_image()
        assert isinstance(pil_image, PIL.Image.Image)
        assert pil_image.size == (909, 720)
    finally:
        image.cleanup()



import importlib
import sys
import time
import traceback
from collections.abc import Callable
from pathlib import Path

import ray
import yaml
from ray import serve
from ray.autoscaler.v2.schema import ResourceDemand
from ray.autoscaler.v2.sdk import get_cluster_status
from ray.serve.config import HTTPOptions
from ray.serve.deployment import Application, Deployment
from ray.serve.schema import ApplicationStatusOverview
from rich import print as rprint

from aana.api.api_generation import Endpoint
from aana.api.event_handlers.event_handler import EventHandler
from aana.api.request_handler import RequestHandler
from aana.configs.settings import settings as aana_settings
from aana.exceptions.runtime import (
    DeploymentException,
    EmptyMigrationsException,
    FailedDeployment,
    InsufficientResources,
)
from aana.storage.op import run_alembic_migrations
from aana.utils.core import import_from_path


class AanaSDK:
    """Aana SDK to deploy and manage Aana deployments and endpoints."""

    def __init__(
        self,
        name: str = "app",
        migration_func: Callable | None = None,
        retryable_exceptions: list[Exception, str] | None = None,
    ):
        """Aana SDK to deploy and manage Aana deployments and endpoints.

        Args:
            name (str, optional): The name of the application. Defaults to "app".
            migration_func (Callable | None): The migration function to run. Defaults to None.
            retryable_exceptions (list[Exception, str] | None): The exceptions that can be retried in the task queue.
                                                                Defaults to ['InferenceException', 'ActorDiedError', 'OutOfMemoryError'].
        """
        self.name = name
        self.migration_func = migration_func
        self.endpoints: dict[str, Endpoint] = {}
        self.deployments: dict[str, Deployment] = {}

        if retryable_exceptions is None:
            self.retryable_exceptions = [
                "InferenceException",
                "ActorDiedError",
                "OutOfMemoryError",
            ]
        else:
            self.retryable_exceptions = retryable_exceptions
        # Convert exceptions to string if they are not already
        # to avoid serialization issues
        self.retryable_exceptions = [
            exc if isinstance(exc, str) else exc.__name__
            for exc in self.retryable_exceptions
        ]

        if aana_settings.task_queue.enabled:
            self.add_task_queue(deploy=False)

    def connect(
        self,
        port: int = 8000,
        host: str = "127.0.0.1",
        address: str = "auto",
        dashboard_host: str = "127.0.0.1",
        dashboard_port: int = 8265,
        show_logs: bool = False,
        num_cpus: int | None = None,
        num_gpus: int | None = None,
    ) -> "AanaSDK":
        """Connect to a Ray cluster or start a new Ray cluster and Ray Serve.

        Args:
            port (int, optional): The port to run the Aana server on. Defaults to 8000.
            host (str, optional): The host to run the Aana server on. Defaults to "127.0.0.1".
            address (str, optional): The address of the Ray cluster. Defaults to "auto".
            dashboard_host (str, optional):  The host to bind the dashboard server to. Can either be
                localhost (127.0.0.1) or 0.0.0.0 (available from all interfaces).
                By default, this is set to localhost to prevent access from external machines.
            dashboard_port (int, optional): The port to bind the dashboard server to. Defaults to 8265.
            show_logs (bool, optional): If True, the logs will be shown, otherwise
                they will be hidden but can be accessed in the Ray dashboard. Defaults to False.
            num_cpus (int, optional): Number of CPUs the user wishes to assign to each
                raylet. By default, this is set based on virtual cores.
            num_gpus (int, optional): Number of GPUs the user wishes to assign to each
                raylet. By default, this is set based on detected GPUs.
        """
        self.port = port
        self.host = host

        try:
            # Try to connect to an existing Ray cluster
            ray.init(
                address=address,
                ignore_reinit_error=True,
                log_to_driver=show_logs,
            )
        except ConnectionError:
            # If connection fails, start a new Ray cluster and serve instance
            ray.init(
                ignore_reinit_error=True,
                log_to_driver=show_logs,
                num_cpus=num_cpus,
                num_gpus=num_gpus,
                include_dashboard=True,
                dashboard_host=dashboard_host,
                dashboard_port=dashboard_port,
            )

        serve_status = serve.status()
        if serve_status.proxies == {}:  # If serve is not running yet
            # TODO: check if the port is already in use if serve is not running yet or
            # check if the port is the same as an existing serve instance if serve is running
            serve.start(http_options=HTTPOptions(port=self.port, host=self.host))

        return self

    def migrate(self):
        """Run Alembic migrations."""
        if self.migration_func:
            try:
                self.migration_func(aana_settings)
            except EmptyMigrationsException:
                print(
                    "No versions found in the custom migrations. Using default migrations."
                )
                run_alembic_migrations(aana_settings)
        else:
            run_alembic_migrations(aana_settings)

    def print_app_status(self, app_name: str, app_status: ApplicationStatusOverview):
        """Show the status of the application using simple ASCII formatting.

        Args:
            app_name (str): The name of the application.
            app_status (ApplicationStatusOverview): The status of the application.
        """

        def print_separator(end="\n"):
            print("=" * 60, end=end)

        def print_header(title):
            print_separator()
            print(title)
            print_separator()

        def print_key_value(key, value, indent=0):
            print(f"{' ' * indent}{key}: {value}")

        if app_status.deployments:
            for deployment_name, deployment_status in app_status.deployments.items():
                print_header(f"{deployment_name} ({app_name})")
                print_key_value("Status", deployment_status.status.value, indent=0)
                print_key_value("Message", deployment_status.message, indent=0)
        print_separator()

    def add_task_queue(self, deploy: bool = False):
        """Add a task queue deployment.

        Args:
            deploy (bool, optional): If True, the deployment will be deployed immediately,
                    otherwise it will be registered and can be deployed later when deploy() is called. Defaults to False.
        """
        from aana.deployments.task_queue_deployment import (
            TaskQueueConfig,
            TaskQueueDeployment,
        )

        task_queue_deployment = TaskQueueDeployment.options(
            num_replicas=1,
            user_config=TaskQueueConfig(
                app_name=self.name,
                retryable_exceptions=self.retryable_exceptions,
            ).model_dump(mode="json"),
        )
        self.register_deployment(
            "task_queue_deployment",
            task_queue_deployment,
            deploy=deploy,
        )

    def register_deployment(
        self,
        name: str,
        instance: Deployment,
        deploy: bool = False,
    ):
        """Register a deployment.

        Args:
            name (str): The name of the deployment.
            instance (Deployment): The instance of the deployment to be registered.
            deploy (bool, optional): If True, the deployment will be deployed immediately,
                    otherwise it will be registered and can be deployed later when deploy() is called. Defaults to False.
        """
        if deploy:
            try:
                serve.api._run(
                    instance.bind(),
                    name=name,
                    route_prefix=f"/{name}",
                    _blocking=False,
                )
                self.wait_for_deployment()
            except FailedDeployment:
                status = serve.status()
                app_status = status.applications[name]
                self.print_app_status(name, app_status)
        else:
            self.deployments[name] = instance

    def get_deployment_app(self, name: str) -> Application:
        """Get the application instance for the deployment.

        Args:
            name (str): The name of the deployment.

        Returns:
            Application: The application instance for the deployment.

        Raises:
            KeyError: If the deployment is not found.
        """
        if name in self.deployments:
            return self.deployments[name].bind()
        else:
            raise KeyError(f"Deployment {name} not found.")  # noqa: TRY003

    def unregister_deployment(self, name: str):
        """Unregister a deployment.

        Args:
            name (str): The name of the deployment to be unregistered.
        """
        if name in self.deployments:
            del self.deployments[name]
        serve.delete(name)

    def get_main_app(self) -> Application:
        """Get the main application instance.

        Returns:
            Application: The main application instance.
        """
        return RequestHandler.options(num_replicas=aana_settings.num_workers).bind(
            app_name=self.name,
            endpoints=self.endpoints.values(),
            deployments=list(self.deployments.keys()),
        )

    def register_endpoint(
        self,
        name: str,
        path: str,
        summary: str,
        endpoint_cls: type[Endpoint],
        event_handlers: list[EventHandler] | None = None,
    ):
        """Register an endpoint.

        Args:
            name (str): The name of the endpoint.
            path (str): The path of the endpoint.
            summary (str): The summary of the endpoint.
            endpoint_cls (Type[Endpoint]): The class of the endpoint.
            event_handlers (list[EventHandler], optional): The event handlers to register for the endpoint.
        """
        endpoint = endpoint_cls(
            name=name,
            path=path,
            summary=summary,
            event_handlers=event_handlers,
        )
        self.endpoints[name] = endpoint

    def unregister_endpoint(self, name: str):
        """Unregister an endpoint.

        Args:
            name (str): The name of the endpoint to be unregistered.
        """
        if name in self.endpoints:
            del self.endpoints[name]

    def wait_for_deployment(self):  # noqa: C901
        """Wait for the deployment to complete."""
        consecutive_resource_unavailable = 0
        # Number of consecutive checks before raising an resource unavailable error
        resource_unavailable_threshold = 5

        while True:
            status = serve.status()
            if all(
                application.status == "RUNNING"
                for application in status.applications.values()
            ):
                break
            if any(
                application.status == "DEPLOY_FAILED"
                or application.status == "UNHEALTHY"
                for application in status.applications.values()
            ):
                error_messages = []
                for app_name, app_status in status.applications.items():
                    if (
                        app_status.status == "DEPLOY_FAILED"
                        or app_status.status == "UNHEALTHY"
                    ):
                        for (
                            deployment_name,
                            deployment_status,
                        ) in app_status.deployments.items():
                            error_messages.append(
                                f"Error: {deployment_name} ({app_name}): {deployment_status.message}"
                            )
                raise FailedDeployment("\n".join(error_messages))

            gcs_address = ray.get_runtime_context().gcs_address
            cluster_status = get_cluster_status(gcs_address)
            demands = (
                cluster_status.resource_demands.cluster_constraint_demand
                + cluster_status.resource_demands.ray_task_actor_demand
                + cluster_status.resource_demands.placement_group_demand
            )

            resource_unavailable = False
            for demand in demands:
                if isinstance(demand, ResourceDemand) and demand.bundles_by_count:
                    error_message = f"Error: No available node types can fulfill resource request {demand.bundles_by_count[0].bundle}. "
                    if "GPU" in demand.bundles_by_count[0].bundle:
                        error_message += "Might be due to insufficient or misconfigured CPU or GPU resources."
                    resource_unavailable = True
                else:
                    error_message = f"Error: {demand}"
                    resource_unavailable = True

            if resource_unavailable:
                consecutive_resource_unavailable += 1
                if consecutive_resource_unavailable >= resource_unavailable_threshold:
                    raise InsufficientResources(error_message)
            else:
                consecutive_resource_unavailable = 0

            time.sleep(1)  # Wait for 1 second before checking again

    def deploy(self, blocking: bool = False, sequential: bool = False):
        """Deploy the application with the registered endpoints and deployments.

        Args:
            blocking (bool, optional): If True, the function will block until interrupted. Defaults to False.
            sequential (bool, optional): If True, the deployments will be deployed sequentially. Defaults to False.
        """
        try:
            for deployment_name in self.deployments:
                serve.api._run(
                    self.get_deployment_app(deployment_name),
                    name=deployment_name,
                    route_prefix=f"/{deployment_name}",
                    _blocking=False,
                )
                if sequential:
                    self.wait_for_deployment()

            serve.api._run(
                self.get_main_app(),
                name=self.name,
                route_prefix="/",
                _blocking=False,  # blocking manually after to display the message "Deployed successfully."
            )

            self.wait_for_deployment()

            rprint("[green]Deployed successfully.[/green]")
            rprint(
                f"Documentation is available at "
                f"[link=http://{self.host}:{self.port}/docs]http://{self.host}:{self.port}/docs[/link] and "
                f"[link=http://{self.host}:{self.port}/redoc]http://{self.host}:{self.port}/redoc[/link]"
            )
            while blocking:
                time.sleep(10)
        except KeyboardInterrupt:
            print("Got KeyboardInterrupt, shutting down...")
            serve.shutdown()
            sys.exit()
        except DeploymentException as e:
            status = serve.status()
            serve.shutdown()
            for app_name, app_status in status.applications.items():
                if (
                    app_status.status == "DEPLOY_FAILED"
                    or app_status.status == "UNHEALTHY"
                ):
                    self.print_app_status(app_name, app_status)
            if isinstance(e, InsufficientResources):
                rprint(f"[red] {e} [/red]")
            raise
        except Exception:
            serve.shutdown()
            traceback.print_exc()
            print(
                "Received unexpected error, see console logs for more details. "
                "Shutting down..."
            )
            raise

    def shutdown(self):
        """Shutdown the Aana server."""
        serve.shutdown()
        ray.shutdown()

    def build(
        self,
        import_path: str,
        host: str = "0.0.0.0",  # noqa: S104
        port: int = 8000,
        app_config_name: str = "app_config",
        config_name: str = "config",
    ):
        """Build the application configuration file.

        Two files will be created: app_config (.py) and config (.yaml).s

        Args:
            import_path (str): The import path of the application.
            host (str): The host to run the application on. Defaults to "0.0.0.0".
            port (int): The port to run the application on. Defaults to 8000.
            app_config_name (str): The name of the application config file. Defaults to "app_config".
            config_name (str): The name of the config file. Defaults to "config".
        """
        # Split the import path into module and variable.
        # For example, aana.projects.whisper.app:aana_app will be split into
        # module "aana.projects.whisper.app" and variable "aana_app".
        app_module, app_var = import_path.split(":")

        # Use location of the app module as the output directory
        output_dir = Path(importlib.util.find_spec(app_module).origin).parent

        # Import AanaSDK app from the given import path
        aana_app = import_from_path(import_path)
        if not isinstance(aana_app, AanaSDK):
            raise TypeError(  # noqa: TRY003
                f"Error: {import_path} is not an AanaSDK instance, got {type(aana_app)}"
            )

        # Generate the app config file
        # Example:
        #    from aana.projects.whisper.app import aana_app
        #    asr_deployment = aana_app.get_deployment_app("asr_deployment")
        #    vad_deployment = aana_app.get_deployment_app("vad_deployment")
        #    whisper_app = aana_app.get_main_app()
        app_config = ""
        app_config += f"from {app_module} import {app_var}\n\n"
        for deployment_name in aana_app.deployments:
            app_config += f"{deployment_name} = {app_var}.get_deployment_app('{deployment_name}')\n"
        app_config += f"{aana_app.name} = {app_var}.get_main_app()\n"

        # Output path for the app config file
        app_config_path = output_dir / f"{app_config_name}.py"
        # Import path for the app config file, for example aana.projects.whisper.app_config
        app_config_import_path = f"{app_module.rsplit('.', 1)[0]}.{app_config_name}"

        # Write the app config file
        with app_config_path.open("w") as f:
            f.write(app_config)

        # Build "serve build" command to generate config.yaml
        # For example,
        # serve build aana.projects.whisper.app_config:vad_deployment
        #             aana.projects.whisper.app_config:asr_deployment
        #             aana.projects.whisper.app_config:whisper_app
        #             -o /workspaces/aana_sdk/aana/projects/whisper/config.yaml
        config_path = output_dir / f"{config_name}.yaml"
        serve_options = []
        for deployment_name in aana_app.deployments:
            serve_options.append(f"{app_config_import_path}:{deployment_name}")  # noqa: PERF401
        serve_options += [
            f"{app_config_import_path}:{aana_app.name}",
            "--output-path",
            str(config_path),
            "--app-dir",
            output_dir,
        ]

        # Execute "serve build" with click CliRuuner
        from click.testing import CliRunner
        from ray.serve.scripts import ServeDeploySchemaDumper, build

        result = CliRunner().invoke(build, serve_options)
        if result.exception:
            raise result.exception

        # Update the config file with the host and port and rename apps
        with config_path.open() as f:
            config = yaml.load(f, Loader=yaml.FullLoader)  # noqa: S506

        config["http_options"] = {"host": host, "port": port}

        for app in config["applications"]:
            app["name"] = app["import_path"].split(":")[-1]

        with config_path.open("w") as f:
            yaml.dump(
                config,
                f,
                Dumper=ServeDeploySchemaDumper,
                default_flow_style=False,
                sort_keys=False,
            )

        print(f"App config successfully saved to {app_config_path}")
        print(f"Config successfully saved to {config_path}")



from collections import defaultdict
from collections.abc import MutableMapping

from aana.api.event_handlers.event_handler import EventHandler
from aana.exceptions.runtime import (
    HandlerNotRegisteredException,
)


class EventManager:
    """Class for event manager. Not guaranteed to be thread safe."""

    _handlers: MutableMapping[str, EventHandler]

    def __init__(self):
        """Constructor."""
        self._handlers = defaultdict(list)

    def handle(self, event_name: str, *args, **kwargs):
        """Trigger event handlers for `event_name`.

        Arguments:
            event_name (str): name of event
            *args (list): specific args
            **kwargs (dict): specific args
        """
        for handler in self._handlers[event_name]:
            handler.handle(event_name, *args, **kwargs)

    def register_handler_for_events(
        self, handler: EventHandler, event_names: list[str]
    ):
        """Adds a handler to the event handler list.

        Arguments:
            handler (EventHandler): the handler to deregister
            event_names (list[str]): the events from which this handler is to be deregistered
        """
        for event_name in event_names:
            if handler not in self._handlers[event_name]:
                self._handlers[event_name].append(handler)

    def deregister_handler_from_event(self, handler: EventHandler, event_name: str):
        """Removes a handler from the event handler list.

        Arguments:
            handler (EventHandler): the handler to remove
            event_name (str): the name of the event from which the handler should be removed
        Raises:
            HandlerNotRegisteredException: if the handler isn't registered. (embed in try-except to suppress)
        """
        try:
            self._handlers[event_name].remove(handler)
        except ValueError as e:
            raise HandlerNotRegisteredException() from e

    def deregister_handler_from_all_events(self, handler: EventHandler):
        """Removes a handler from all event handlers.

        Arguments:
            handler (EventHandler): the exact instance of the handler to remove.

        Raises:
            HandlerNotRegisteredException: if the handler isn't registered. (embed in try-except to suppress)
        """
        has_removed = False
        for handler_list in self._handlers.values():
            if handler in handler_list:
                handler_list.remove(handler)
                has_removed = True
        if not has_removed:
            raise HandlerNotRegisteredException()



"""init.

Revision ID: 5ad873484aa3
Revises: 
Create Date: 2024-07-25 13:09:44.450321

"""
from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "5ad873484aa3"
down_revision: str | None = None
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade database to this revision from previous."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "caption",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column(
            "model",
            sa.String(),
            nullable=False,
            comment="Name of model used to generate the caption",
        ),
        sa.Column(
            "frame_id",
            sa.Integer(),
            nullable=False,
            comment="The 0-based frame id of video for caption",
        ),
        sa.Column("caption", sa.String(), nullable=False, comment="Frame caption"),
        sa.Column(
            "timestamp",
            sa.Float(),
            nullable=False,
            comment="Frame timestamp in seconds",
        ),
        sa.Column(
            "caption_type", sa.String(), nullable=False, comment="The type of caption"
        ),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
            comment="Timestamp when row is inserted",
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
            comment="Timestamp when row is updated",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_caption")),
    )
    op.create_table(
        "media",
        sa.Column(
            "id",
            sa.String(length=36),
            nullable=False,
            comment="Unique identifier for the media",
        ),
        sa.Column(
            "media_type", sa.String(), nullable=False, comment="The type of media"
        ),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
            comment="Timestamp when row is inserted",
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
            comment="Timestamp when row is updated",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_media")),
    )
    op.create_table(
        "tasks",
        sa.Column("id", sa.UUID(), nullable=False, comment="Task ID"),
        sa.Column(
            "endpoint",
            sa.String(),
            nullable=False,
            comment="The endpoint to which the task is assigned",
        ),
        sa.Column("data", sa.PickleType(), nullable=False, comment="Data for the task"),
        sa.Column(
            "status",
            sa.Enum(
                "CREATED",
                "ASSIGNED",
                "COMPLETED",
                "RUNNING",
                "FAILED",
                "NOT_FINISHED",
                name="status",
            ),
            nullable=False,
            comment="Status of the task",
        ),
        sa.Column(
            "priority",
            sa.Integer(),
            nullable=False,
            comment="Priority of the task (0 is the lowest)",
        ),
        sa.Column(
            "assigned_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=True,
            comment="Timestamp when the task was assigned",
        ),
        sa.Column(
            "completed_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=True,
            comment="Timestamp when the task was completed",
        ),
        sa.Column(
            "progress",
            sa.Float(),
            nullable=False,
            comment="Progress of the task in percentage",
        ),
        sa.Column(
            "result",
            sa.JSON(),
            nullable=True,
            comment="Result of the task in JSON format",
        ),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
            comment="Timestamp when row is inserted",
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
            comment="Timestamp when row is updated",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_tasks")),
    )
    op.create_table(
        "transcript",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column(
            "model",
            sa.String(),
            nullable=False,
            comment="Name of model used to generate transcript",
        ),
        sa.Column(
            "transcript",
            sa.String(),
            nullable=False,
            comment="Full text transcript of media",
        ),
        sa.Column(
            "segments", sa.JSON(), nullable=False, comment="Segments of the transcript"
        ),
        sa.Column(
            "language",
            sa.String(),
            nullable=False,
            comment="Language of the transcript as predicted by model",
        ),
        sa.Column(
            "language_confidence",
            sa.Float(),
            nullable=False,
            comment="Confidence score of language prediction",
        ),
        sa.Column(
            "transcript_type",
            sa.String(),
            nullable=False,
            comment="The type of transcript",
        ),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
            comment="Timestamp when row is inserted",
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
            comment="Timestamp when row is updated",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_transcript")),
    )
    op.create_table(
        "video",
        sa.Column("id", sa.String(length=36), nullable=False),
        sa.Column("path", sa.String(), nullable=True, comment="Path"),
        sa.Column("url", sa.String(), nullable=True, comment="URL"),
        sa.Column("title", sa.String(), nullable=True, comment="Title"),
        sa.Column("description", sa.String(), nullable=True, comment="Description"),
        sa.ForeignKeyConstraint(["id"], ["media.id"], name=op.f("fk_video_id_media")),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_video")),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database from this revision to previous."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("video")
    op.drop_table("transcript")
    op.drop_table("tasks")
    op.drop_table("media")
    op.drop_table("caption")
    # ### end Alembic commands ###



import asyncio
import threading
from collections.abc import Coroutine
from typing import Any, TypeVar

__all__ = ["run_async"]

T = TypeVar("T")


def run_async(coro: Coroutine[Any, Any, T]) -> T:
    """Run a coroutine in a thread if the current thread is running an event loop.

    Otherwise, run the coroutine in the current asyncio loop.

    Useful when you want to run an async function in a non-async context.

    From: https://stackoverflow.com/a/75094151

    Args:
        coro (Coroutine): The coroutine to run.

    Returns:
        T: The result of the coroutine.
    """

    class RunThread(threading.Thread):
        """Run a coroutine in a thread."""

        def __init__(self, coro: Coroutine[Any, Any, T]):
            """Initialize the thread."""
            self.coro = coro
            self.result: T | None = None
            self.exception: Exception | None = None
            super().__init__()

        def run(self):
            """Run the coroutine."""
            try:
                self.result = asyncio.run(self.coro)
            except Exception as e:
                self.exception = e

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None

    if loop and loop.is_running():
        thread = RunThread(coro)
        thread.start()
        thread.join()
        if thread.exception:
            raise thread.exception
        return thread.result
    else:
        return asyncio.run(coro)



# ruff: noqa: S101, S113
import json
from typing import TypedDict

import requests
from ray import serve

from aana.api.api_generation import Endpoint
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.base_deployment import BaseDeployment


@serve.deployment
class Lowercase(BaseDeployment):
    """Ray deployment that returns the lowercase version of a text."""

    async def lower(self, text: str) -> dict:
        """Lowercase the text.

        Args:
            text (str): The text to lowercase

        Returns:
            dict: The lowercase text
        """
        return {"text": text.lower()}


class LowercaseEndpointOutput(TypedDict):
    """The output of the lowercase endpoint."""

    text: str


class LowercaseEndpoint(Endpoint):
    """Lowercase endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.lowercase_handle = await AanaDeploymentHandle.create(
            "lowercase_deployment"
        )
        await super().initialize()

    async def run(self, text: str) -> LowercaseEndpointOutput:
        """Lowercase the text.

        Args:
            text (TextList): The list of text to lowercase

        Returns:
            LowercaseEndpointOutput: The lowercase texts
        """
        lowercase_output = await self.lowercase_handle.lower(text=text)
        return {"text": lowercase_output["text"]}


deployments = [
    {
        "name": "lowercase_deployment",
        "instance": Lowercase,
    }
]

endpoints = [
    {
        "name": "lowercase",
        "path": "/lowercase",
        "summary": "Lowercase text",
        "endpoint_cls": LowercaseEndpoint,
    }
]


def test_app(create_app):
    """Test the Ray Serve app."""
    aana_app = create_app(deployments, endpoints)

    port = aana_app.port
    route_prefix = ""

    # Check that the server is ready
    response = requests.get(f"http://localhost:{port}{route_prefix}/api/ready")
    assert response.status_code == 200
    assert response.json() == {"ready": True}

    # Test lowercase endpoint
    data = {"text": "Hello World! This is a test."}
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/lowercase",
        data={"body": json.dumps(data)},
    )
    assert response.status_code == 200
    lowercase_text = response.json().get("text")
    assert lowercase_text == "hello world! this is a test."



# ruff: noqa: S101

import pytest
from pydantic import BaseModel

from aana.core.models.base import merged_options


class MyOptions(BaseModel):
    """Test option class."""

    field1: str
    field2: int | None = None
    field3: bool
    field4: str = "default"


def test_merged_options_same_type():
    """Test merged_options with options of the same type as default_options."""
    default = MyOptions(field1="default1", field2=2, field3=True)
    to_merge = MyOptions(field1="merge1", field2=None, field3=False)
    merged = merged_options(default, to_merge)

    assert merged.field1 == "merge1"
    assert (
        merged.field2 == 2
    )  # Should retain value from default_options as it's None in options
    assert merged.field3 == False


def test_merged_options_none():
    """Test merged_options with options=None."""
    default = MyOptions(field1="default1", field2=2, field3=True)
    merged = merged_options(default, None)

    assert merged.model_dump() == default.model_dump()


def test_merged_options_type_mismatch():
    """Test merged_options with options of a different type from default_options."""

    class AnotherOptions(BaseModel):
        another_field: str

    default = MyOptions(field1="default1", field2=2, field3=True)
    to_merge = AnotherOptions(another_field="test")

    with pytest.raises(ValueError):
        merged_options(default, to_merge)


def test_merged_options_unset():
    """Test merged_options with unset fields."""
    default = MyOptions(field1="default1", field2=2, field3=True, field4="new_default")
    to_merge = MyOptions(field1="merge1", field3=False)  # field4 is not set
    merged = merged_options(default, to_merge)

    assert merged.field1 == "merge1"
    assert merged.field2 == 2
    assert merged.field3 == False
    assert (
        merged.field4 == "new_default"
    )  # Should retain value from default_options as it's not set in options



# ruff: noqa: S101, A003
from collections.abc import AsyncGenerator
from typing import Annotated, TypedDict

import pytest
from pydantic import BaseModel, ConfigDict, Field

from aana.api.api_generation import Endpoint
from aana.exceptions.runtime import MultipleFileUploadNotAllowed


class InputModel(BaseModel):
    """Model for a text input."""

    input: str = Field(..., description="Input text")
    model_config = ConfigDict(extra="forbid")


class FileUploadModel(BaseModel):
    """Model for a file upload input."""

    content: bytes | None = Field(
        None,
        description="The content in bytes. Set this field to 'file' to upload files to the endpoint.",
    )

    def set_files(self, files):
        """Set files."""
        if files:
            if isinstance(files, list):
                files = files[0]
            self.content = files

    model_config = ConfigDict(
        extra="forbid", file_upload=True, file_upload_description="Upload image files."
    )


class TestEndpointOutput(TypedDict):
    """The output of the test endpoint."""

    output: Annotated[str, Field(..., description="Output text")]


class TestEndpoint(Endpoint):
    """Test endpoint for get_request_model."""

    async def run(self, input_data: InputModel) -> TestEndpointOutput:
        """Run the endpoint."""
        return {"output": input_data.input}


class TestFileUploadEndpoint(Endpoint):
    """Test endpoint for __get_file_upload_field."""

    async def run(self, input_data: FileUploadModel) -> TestEndpointOutput:
        """Run the endpoint."""
        return {"output": "file uploaded"}


class TestMultipleFileUploadEndpoint(Endpoint):
    """Test endpoint for __get_file_upload_field with multiple file uploads."""

    async def run(
        self, input_data: FileUploadModel, input_data2: FileUploadModel
    ) -> TestEndpointOutput:
        """Run the endpoint."""
        return {"output": "file uploaded"}


class TestEndpointMissingReturn(Endpoint):
    """Test endpoint for get_response_model with missing return type."""

    async def run(self, input_data: InputModel):
        """Run the endpoint."""
        return {"output": input_data.input}


class TestEndpointMissingReturnStreaming(Endpoint):
    """Test endpoint for get_response_model with missing return type in streaming endpoint."""

    async def run(self, input_data: InputModel) -> AsyncGenerator:
        """Run the endpoint."""
        yield {"output": input_data.input}


def test_get_request_model():
    """Test the get_request_model function."""
    endpoint = TestEndpoint(
        name="test_endpoint",
        summary="Test endpoint",
        path="/test_endpoint",
    )

    RequestModel = endpoint.get_request_model()

    # Check that the request model named correctly
    assert RequestModel.__name__ == "TestEndpointRequest"

    # Check that the request model has the correct fields
    assert RequestModel.model_fields.keys() == {"input_data"}

    # Check that the request fields have the correct types
    assert RequestModel.model_fields["input_data"].annotation == InputModel


def test_get_response_model():
    """Test the get_response_model function."""
    endpoint = TestEndpoint(
        name="test_endpoint",
        summary="Test endpoint",
        path="/test_endpoint",
    )
    ResponseModel = endpoint.get_response_model()

    # Check that the response model named correctly
    assert ResponseModel.__name__ == "TestEndpointResponse"

    # Check that the response model has the correct fields
    assert ResponseModel.model_fields.keys() == {"output"}

    # Check that the response fields have the correct types
    assert ResponseModel.model_fields["output"].annotation == str


def test_get_file_upload_field():
    """Test the __get_file_upload_field function."""
    endpoint = TestFileUploadEndpoint(
        name="test_endpoint",
        summary="Test endpoint",
        path="/test_endpoint",
    )

    file_upload_field = endpoint._Endpoint__get_file_upload_field()

    # Check that the file upload field named correctly
    assert file_upload_field.name == "input_data"

    # Check that the file upload field has the correct description
    assert file_upload_field.description == "Upload image files."


def test_get_file_upload_field_multiple_file_uploads():
    """Test the __get_file_upload_field function with multiple file uploads."""
    endpoint = TestMultipleFileUploadEndpoint(
        name="test_endpoint",
        summary="Test endpoint",
        path="/test_endpoint",
    )

    with pytest.raises(MultipleFileUploadNotAllowed):
        endpoint._Endpoint__get_file_upload_field()


def test_get_response_model_missing_return():
    """Test the get_response_model function with missing return type."""
    endpoint = TestEndpointMissingReturn(
        name="test_endpoint",
        summary="Test endpoint",
        path="/test_endpoint",
    )

    with pytest.raises(ValueError):
        endpoint.get_response_model()


def test_get_response_model_missing_return_streaming():
    """Test the get_response_model function with missing return type in streaming endpoint."""
    endpoint = TestEndpointMissingReturnStreaming(
        name="test_endpoint",
        summary="Test endpoint",
        path="/test_endpoint",
    )

    with pytest.raises(ValueError):
        endpoint.get_response_model()



from typing import Annotated

import numpy as np
from faster_whisper.transcribe import (
    Segment as WhisperSegment,
)
from faster_whisper.transcribe import (
    TranscriptionInfo as WhisperTranscriptionInfo,
)
from faster_whisper.transcribe import (
    Word as WhisperWord,
)
from pydantic import BaseModel, ConfigDict, Field

from aana.core.models.time import TimeInterval

__all__ = [
    "AsrWord",
    "AsrSegment",
    "AsrTranscriptionInfo",
    "AsrTranscription",
    "AsrSegments",
    "AsrSegmentsList",
    "AsrTranscriptionInfoList",
    "AsrTranscriptionList",
]


class AsrWord(BaseModel):
    """Pydantic schema for Word from ASR model.

    Attributes:
        word (str): The word text.
        speaker (str| None): Speaker label for the word.
        time_interval (TimeInterval): Time interval of the word.
        alignment_confidence (float): Alignment confidence of the word, >= 0.0 and <= 1.0.
    """

    word: str = Field(description="The word text")
    speaker: str | None = Field(None, description="Speaker label for the word")
    time_interval: TimeInterval = Field(description="Time interval of the word")
    alignment_confidence: float = Field(
        ge=0.0, le=1.0, description="Alignment confidence of the word"
    )

    @classmethod
    def from_whisper(cls, whisper_word: WhisperWord) -> "AsrWord":
        """Convert WhisperWord to AsrWord.

        Args:
            whisper_word (WhisperWord): The WhisperWord from faster-whisper.

        Returns:
            AsrWord: The converted AsrWord.
        """
        return cls(
            speaker=None,
            word=whisper_word.word,
            time_interval=TimeInterval(start=whisper_word.start, end=whisper_word.end),
            alignment_confidence=whisper_word.probability,
        )

    model_config = ConfigDict(
        json_schema_extra={
            "description": "Word",
        }
    )


class AsrSegment(BaseModel):
    """Pydantic schema for Segment from ASR model.

    Attributes:
        text (str): The text of the segment (transcript/translation).
        time_interval (TimeInterval): Time interval of the segment.
        confidence (float | None): Confidence of the segment.
        no_speech_confidence (float | None): Chance of being a silence segment.
        words (list[AsrWord]): List of words in the segment. Default is [].
        speaker (str | None): Speaker label. Default is None.
    """

    text: str = Field(description="The text of the segment (transcript/translation)")
    time_interval: TimeInterval = Field(description="Time interval of the segment")
    confidence: float | None = Field(
        None, ge=0.0, le=1.0, description="Confidence of the segment"
    )
    no_speech_confidence: float | None = Field(
        None, ge=0.0, le=1.0, description="Chance of being a silence segment"
    )
    words: list[AsrWord] = Field(
        description="List of words in the segment", default_factory=list
    )
    speaker: str | None = Field(None, description="speaker label of the segment")

    @classmethod
    def from_whisper(cls, whisper_segment: WhisperSegment) -> "AsrSegment":
        """Convert WhisperSegment to AsrSegment."""
        time_interval = TimeInterval(
            start=whisper_segment.start, end=whisper_segment.end
        )
        try:
            avg_logprob = whisper_segment.avg_logprob
            confidence = np.exp(avg_logprob)
        except AttributeError:
            confidence = None

        try:
            words = [AsrWord.from_whisper(word) for word in whisper_segment.words]
        except TypeError:  # "None type object is not iterable"
            words = []
        except AttributeError:  # "'StreamSegment' object has no attribute 'words'"
            words = []
        try:
            no_speech_confidence = whisper_segment.no_speech_prob
        except AttributeError:
            no_speech_confidence = None

        return cls(
            text=whisper_segment.text,
            time_interval=time_interval,
            confidence=confidence,
            no_speech_confidence=no_speech_confidence,
            words=words,
            speaker=None,
        )

    model_config = ConfigDict(
        json_schema_extra={
            "description": "Segment",
        }
    )


class AsrTranscriptionInfo(BaseModel):
    """Pydantic schema for TranscriptionInfo.

    Attributes:
        language (str): Language of the transcription.
        language_confidence (float): Confidence of the language detection, >= 0.0 and <= 1.0. Default is 0.0.
    """

    language: str = Field(description="Language of the transcription", default="")
    language_confidence: float = Field(
        ge=0.0, le=1.0, description="Confidence of the language detection", default=0.0
    )

    @classmethod
    def from_whisper(
        cls, transcription_info: WhisperTranscriptionInfo
    ) -> "AsrTranscriptionInfo":
        """Convert WhisperTranscriptionInfo to AsrTranscriptionInfo.

        Args:
            transcription_info (WhisperTranscriptionInfo): The WhisperTranscriptionInfo from faster-whisper.

        Returns:
            AsrTranscriptionInfo: The converted AsrTranscriptionInfo.
        """
        return cls(
            language=transcription_info.language,
            language_confidence=transcription_info.language_probability,
        )

    def __add__(self, other: "AsrTranscriptionInfo") -> "AsrTranscriptionInfo":
        """Sum two transcriptions info."""
        if self.language == other.language:
            # if the languages are the same, take the average of the confidence
            language = self.language
            language_confidence = (
                self.language_confidence + other.language_confidence
            ) / 2
        else:
            # if the languages are different, take the one with the highest confidence
            if self.language_confidence > other.language_confidence:
                language = self.language
                language_confidence = self.language_confidence
            else:
                language = other.language
                language_confidence = other.language_confidence
        return AsrTranscriptionInfo(
            language=language, language_confidence=language_confidence
        )

    model_config = ConfigDict(
        json_schema_extra={
            "description": "Transcription info",
        }
    )


class AsrTranscription(BaseModel):
    """Pydantic schema for Transcription/Translation.

    Attributes:
        text (str): The text of the transcription/translation. Default is "".
    """

    text: str = Field(
        description="The text of the transcription/translation", default=""
    )

    def __add__(self, other: "AsrTranscription") -> "AsrTranscription":
        """Sum two transcriptions."""
        if self.text == "":
            text = other.text
        elif other.text == "":
            text = self.text
        else:
            text = self.text + "\n" + other.text
        return AsrTranscription(text=text)

    model_config = ConfigDict(
        json_schema_extra={
            "description": "Transcription/Translation",
        }
    )


AsrSegments = Annotated[
    list[AsrSegment], Field(..., description="List of ASR segments")
]

AsrSegmentsList = Annotated[
    list[AsrSegments], Field(..., description="List of lists of ASR segments")
]
AsrTranscriptionInfoList = Annotated[
    list[AsrTranscriptionInfo], Field(..., description="List of ASR transcription info")
]
AsrTranscriptionList = Annotated[
    list[AsrTranscription], Field(..., description="List of ASR transcription")
]



from fastapi import FastAPI
from pydantic import ValidationError

from aana.api.exception_handler import (
    aana_exception_handler,
    validation_exception_handler,
)

app = FastAPI()

app.add_exception_handler(ValidationError, validation_exception_handler)
app.add_exception_handler(Exception, aana_exception_handler)



# ruff: noqa: S101
import json
from importlib import resources
from pathlib import Path
from typing import Literal

import pytest

from aana.core.models.asr import AsrSegment
from aana.core.models.speaker import SpeakerDiarizationSegment
from aana.processors.speaker import PostProcessingForDiarizedAsr
from aana.tests.utils import verify_deployment_results


@pytest.mark.parametrize("audio_file", ["sd_sample.wav"])
def test_asr_diarization_post_process(audio_file: Literal["sd_sample.wav"]):
    """Test that the ASR output can be processed to generate diarized transcription and an invalid ASR output leads to ValueError."""
    # Load precomputed ASR and Diarization outputs
    asr_path = (
        resources.files("aana.tests.files.expected.whisper")
        / f"whisper_medium_{audio_file}.json"
    )
    diar_path = (
        resources.files("aana.tests.files.expected.sd")
        / f"{Path(audio_file).stem}.json"
    )
    expected_results_path = (
        resources.files("aana.tests.files.expected.whisper")
        / f"whisper_medium_{audio_file}_diar.json"
    )

    # convert to WhisperOutput and SpeakerDiarizationOutput
    with Path.open(asr_path, "r") as json_file:
        asr_op = json.load(json_file)

    asr_segments = [
        AsrSegment.model_validate(segment) for segment in asr_op["segments"]
    ]

    with Path.open(diar_path, "r") as json_file:
        diar_op = json.load(json_file)

    diarized_segments = [
        SpeakerDiarizationSegment.model_validate(segment)
        for segment in diar_op["segments"]
    ]
    asr_op["segments"] = PostProcessingForDiarizedAsr.process(
        diarized_segments=diarized_segments, transcription_segments=asr_segments
    )
    verify_deployment_results(expected_results_path, asr_op)

    # Raise error if the ASR output is a invalid input for combining with diarization.

    # setting words to empty list
    for segment in asr_segments:
        segment.words = []

    # Expect ValueError with the specific error message
    with pytest.raises(
        ValueError, match="Word-level timestamps are required for diarized ASR."
    ):
        PostProcessingForDiarizedAsr.process(
            diarized_segments=diarized_segments, transcription_segments=asr_segments
        )



import json
import time
from typing import Annotated, Any
from uuid import UUID, uuid4

import orjson
import ray
from fastapi import Depends
from fastapi.openapi.utils import get_openapi
from fastapi.responses import StreamingResponse
from ray import serve
from sqlalchemy.orm import Session

from aana.api.api_generation import Endpoint, add_custom_schemas_to_openapi_schema
from aana.api.app import app
from aana.api.event_handlers.event_manager import EventManager
from aana.api.exception_handler import custom_exception_handler
from aana.api.responses import AanaJSONResponse
from aana.configs.settings import settings as aana_settings
from aana.core.models.api import DeploymentStatus, SDKStatus, SDKStatusResponse
from aana.core.models.chat import ChatCompletion, ChatCompletionRequest, ChatDialog
from aana.core.models.sampling import SamplingParams
from aana.core.models.task import TaskId, TaskInfo
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.storage.models.task import Status as TaskStatus
from aana.storage.repository.task import TaskRepository
from aana.storage.session import get_session


def get_db():
    """Get a database session."""
    db = get_session()
    try:
        yield db
    finally:
        db.close()


@serve.deployment(ray_actor_options={"num_cpus": 0.1})
@serve.ingress(app)
class RequestHandler:
    """This class is used to handle requests to the Aana application."""

    ready = False

    def __init__(
        self, app_name: str, endpoints: list[Endpoint], deployments: list[str]
    ):
        """Constructor.

        Args:
            app_name (str): The name of the application.
            endpoints (dict): List of endpoints for the request.
            deployments (list[str]): List of deployment names for the app.
        """
        self.app_name = app_name
        self.endpoints = endpoints
        self.deployments = deployments

        self.event_manager = EventManager()
        self.custom_schemas: dict[str, dict] = {}
        for endpoint in self.endpoints:
            endpoint.register(
                app=app,
                custom_schemas=self.custom_schemas,
                event_manager=self.event_manager,
            )

        app.openapi = self.custom_openapi
        self.ready = True
        self.running_tasks = set()

    def custom_openapi(self) -> dict[str, Any]:
        """Returns OpenAPI schema, generating it if necessary."""
        if app.openapi_schema:
            return app.openapi_schema
        # TODO: populate title and version from package info
        openapi_schema = get_openapi(title="Aana", version="0.1.0", routes=app.routes)
        openapi_schema = add_custom_schemas_to_openapi_schema(
            openapi_schema=openapi_schema, custom_schemas=self.custom_schemas
        )
        app.openapi_schema = openapi_schema
        return app.openapi_schema

    @app.get("/api/ready")
    async def is_ready(self):
        """The endpoint for checking if the application is ready.

        Real reason for this endpoint is to make automatic endpoint generation work.
        If RequestHandler doesn't have any endpoints defined manually,
        then the automatic endpoint generation doesn't work.
        #TODO: Find a better solution for this.

        Returns:
            AanaJSONResponse: The response containing the ready status.
        """
        return AanaJSONResponse(content={"ready": self.ready})

    async def check_health(self):
        """Check the health of the application."""
        # Heartbeat for the running tasks
        with get_session() as session:
            task_repo = TaskRepository(session)
            task_repo.heartbeat(self.running_tasks)

    async def execute_task(self, task_id: str | UUID) -> Any:
        """Execute a task.

        Args:
            task_id (str | UUID): The ID of the task.

        Returns:
            Any: The response from the endpoint.
        """
        try:
            self.running_tasks.add(task_id)
            with get_session() as session:
                task_repo = TaskRepository(session)
                task = task_repo.read(task_id)
                path = task.endpoint
                kwargs = task.data

                task_repo.update_status(task_id, TaskStatus.RUNNING, 0)

            for e in self.endpoints:
                if e.path == path:
                    endpoint = e
                    break
            else:
                raise ValueError(f"Endpoint {path} not found")  # noqa: TRY003, TRY301

            if not endpoint.initialized:
                await endpoint.initialize()

            if endpoint.is_streaming_response():
                out = [item async for item in endpoint.run(**kwargs)]
            else:
                out = await endpoint.run(**kwargs)

            with get_session() as session:
                TaskRepository(session).update_status(
                    task_id, TaskStatus.COMPLETED, 100, out
                )
        except Exception as e:
            error_response = custom_exception_handler(None, e)
            error = orjson.loads(error_response.body)
            with get_session() as session:
                TaskRepository(session).update_status(
                    task_id, TaskStatus.FAILED, 0, error
                )
        else:
            return out
        finally:
            self.running_tasks.remove(task_id)

    @app.get(
        "/tasks/get/{task_id}",
        summary="Get Task Status",
        description="Get the task status by task ID.",
        include_in_schema=aana_settings.task_queue.enabled,
    )
    async def get_task_endpoint(
        self, task_id: str, db: Annotated[Session, Depends(get_db)]
    ) -> TaskInfo:
        """Get the task with the given ID.

        Args:
            task_id (str): The ID of the task.
            db (Session): The database session.

        Returns:
            TaskInfo: The status of the task.
        """
        task_repo = TaskRepository(db)
        task = task_repo.read(task_id)
        return TaskInfo(
            id=str(task.id),
            status=task.status,
            result=task.result,
        )

    @app.get(
        "/tasks/delete/{task_id}",
        summary="Delete Task",
        description="Delete the task by task ID.",
        include_in_schema=aana_settings.task_queue.enabled,
    )
    async def delete_task_endpoint(
        self, task_id: str, db: Annotated[Session, Depends(get_db)]
    ) -> TaskId:
        """Delete the task with the given ID.

        Args:
            task_id (str): The ID of the task.
            db (Session): The database session.

        Returns:
            TaskInfo: The deleted task.
        """
        task_repo = TaskRepository(db)
        task = task_repo.delete(task_id)
        return TaskId(task_id=str(task.id))

    @app.post("/chat/completions", response_model=ChatCompletion)
    async def chat_completions(self, request: ChatCompletionRequest):
        """Handle chat completions requests for OpenAI compatible API."""

        async def _async_chat_completions(
            handle: AanaDeploymentHandle,
            dialog: ChatDialog,
            sampling_params: SamplingParams,
        ):
            async for response in handle.chat_stream(
                dialog=dialog, sampling_params=sampling_params
            ):
                chunk = {
                    "id": f"chatcmpl-{uuid4().hex}",
                    "object": "chat.completion.chunk",
                    "model": request.model,
                    "created": int(time.time()),
                    "choices": [
                        {
                            "index": 0,
                            "delta": {"content": response["text"], "role": "assistant"},
                        }
                    ],
                }
                yield f"data: {json.dumps(chunk)}\n\n"
            yield "data: [DONE]\n\n"

        # Check if the deployment exists
        try:
            handle = await AanaDeploymentHandle.create(request.model)
        except ray.serve.exceptions.RayServeException:
            return AanaJSONResponse(
                content={
                    "error": {"message": f"The model `{request.model}` does not exist."}
                },
                status_code=404,
            )

        # Check if the deployment is a chat model
        if not hasattr(handle, "chat") or not hasattr(handle, "chat_stream"):
            return AanaJSONResponse(
                content={
                    "error": {"message": f"The model `{request.model}` does not exist."}
                },
                status_code=404,
            )

        dialog = ChatDialog(
            messages=request.messages,
        )

        sampling_params = SamplingParams(
            temperature=request.temperature,
            max_tokens=request.max_tokens,
            top_p=request.top_p,
        )

        if request.stream:
            return StreamingResponse(
                _async_chat_completions(handle, dialog, sampling_params),
                media_type="application/x-ndjson",
            )
        else:
            response = await handle.chat(dialog=dialog, sampling_params=sampling_params)
            return {
                "id": f"chatcmpl-{uuid4().hex}",
                "object": "chat.completion",
                "model": request.model,
                "created": int(time.time()),
                "choices": [{"index": 0, "message": response["message"]}],
            }

    @app.get("/api/status", response_model=SDKStatusResponse)
    async def status(self) -> SDKStatusResponse:
        """The endpoint for checking the status of the application."""
        app_names = [
            self.app_name,
            *self.deployments,
        ]  # the list of Ray Serve apps that belong to this Aana app
        serve_status = serve.status()
        app_statuses = {
            app_name: app_status
            for app_name, app_status in serve_status.applications.items()
            if app_name in app_names
        }

        app_status_message = ""
        if any(
            app.status in ["DEPLOY_FAILED", "UNHEALTHY", "NOT_STARTED"]
            for app in app_statuses.values()
        ):
            sdk_status = SDKStatus.UNHEALTHY
            error_messages = []
            for app_name, app_status in app_statuses.items():
                if app_status.status in ["DEPLOY_FAILED", "UNHEALTHY"]:
                    for (
                        deployment_name,
                        deployment_status,
                    ) in app_status.deployments.items():
                        error_messages.append(
                            f"Error: {deployment_name} ({app_name}): {deployment_status.message}"
                        )
            app_status_message = "\n".join(error_messages)
        elif all(app.status == "RUNNING" for app in app_statuses.values()):
            sdk_status = SDKStatus.RUNNING
        elif any(
            app.status in ["DEPLOYING", "DELETING"] for app in app_statuses.values()
        ):
            sdk_status = SDKStatus.DEPLOYING
        else:
            sdk_status = SDKStatus.UNHEALTHY
            app_status_message = "Unknown status"

        deployment_statuses = {}
        for app_name, app_status in app_statuses.items():
            messages = []
            for deployment_name, deployment_status in app_status.deployments.items():
                if deployment_status.message:
                    messages.append(
                        f"{deployment_name} ({app_name}): {deployment_status.message}"
                    )
            message = "\n".join(messages)

            deployment_statuses[app_name] = DeploymentStatus(
                status=app_status.status, message=message
            )

        return SDKStatusResponse(
            status=sdk_status,
            message=app_status_message,
            deployments=deployment_statuses,
        )



from functools import lru_cache
from importlib import resources

from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from aana.core.models.chat import ChatDialog


@lru_cache(maxsize=128)
def load_chat_template(chat_template_name: str) -> str:
    """Loads a chat template from the chat templates directory.

    Args:
        chat_template_name (str): The name of the chat template to load.

    Returns:
        str: The loaded chat template.

    Raises:
        ValueError: If the chat template does not exist.
    """
    path = resources.files("aana.core.chat.templates") / f"{chat_template_name}.jinja"
    if not path.exists():
        raise ValueError(f"Chat template {chat_template_name} does not exist.")  # noqa: TRY003

    return path.read_text()


def apply_chat_template(
    tokenizer: PreTrainedTokenizerBase,
    dialog: ChatDialog | list[dict],
    chat_template_name: str | None = None,
) -> str:
    """Applies a chat template to a list of messages to generate a prompt for the model.

    If the chat template is not specified, the tokenizer's default chat template is used.
    If the chat template is specified, the template with given name loaded from the chat templates directory.

    Args:
        tokenizer (PreTrainedTokenizerBase): The tokenizer to use.
        dialog (ChatDialog | list[dict]): The dialog to generate a prompt for.
        chat_template_name (str, optional): The name of the chat template to use. Defaults to None, which uses the tokenizer's default chat template.

    Returns:
        str: The generated prompt.

    Raises:
        ValueError: If the tokenizer does not have a chat template.
        ValueError: If the chat template does not exist.
    """
    if isinstance(dialog, ChatDialog):
        messages = dialog.model_dump()["messages"]
    else:
        messages = dialog

    if chat_template_name is not None:
        chat_template = load_chat_template(chat_template_name)
        tokenizer.chat_template = chat_template

    if tokenizer.chat_template is None:
        raise ValueError("Tokenizer does not have a chat template.")  # noqa: TRY003

    return tokenizer.apply_chat_template(
        messages, tokenize=False, add_generation_prompt=True
    )



from __future__ import annotations  # Let classes use themselves in type annotations

from typing import TYPE_CHECKING

from sqlalchemy import JSON, CheckConstraint
from sqlalchemy.orm import Mapped, mapped_column

from aana.storage.models.base import BaseEntity, TimeStampEntity

if TYPE_CHECKING:
    from aana.core.models.asr import (
        AsrSegments,
        AsrTranscription,
        AsrTranscriptionInfo,
    )


class TranscriptEntity(BaseEntity, TimeStampEntity):
    """ORM class for media transcripts generated by a model.

    Attributes:
        id (int): Unique identifier for the transcript.
        model (str): Name of the model used to generate the transcript.
        transcript (str): Full text transcript of the media.
        segments (dict): Segments of the transcript.
        language (str): Language of the transcript as predicted by the model.
        language_confidence (float): Confidence score of language prediction.
        transcript_type (str): The type of transcript (populated automatically by ORM based on `polymorphic_identity` of subclass).
    """

    __tablename__ = "transcript"

    id: Mapped[int] = mapped_column(autoincrement=True, primary_key=True)
    model: Mapped[str] = mapped_column(
        nullable=False, comment="Name of model used to generate transcript"
    )
    transcript: Mapped[str] = mapped_column(comment="Full text transcript of media")
    segments: Mapped[dict] = mapped_column(JSON, comment="Segments of the transcript")
    language: Mapped[str] = mapped_column(
        comment="Language of the transcript as predicted by model"
    )
    language_confidence: Mapped[float] = mapped_column(
        CheckConstraint(
            "0 <= language_confidence <= 1", name="language_confidence_value_range"
        ),
        comment="Confidence score of language prediction",
    )
    transcript_type: Mapped[str] = mapped_column(comment="The type of transcript")

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "transcript",
        "polymorphic_on": "transcript_type",
    }

    @classmethod
    def from_asr_output(
        cls,
        model_name: str,
        info: AsrTranscriptionInfo,
        transcription: AsrTranscription,
        segments: AsrSegments,
    ) -> TranscriptEntity:
        """Converts an AsrTranscriptionInfo and AsrTranscription to a single Transcript entity.

        Args:
            model_name (str): Name of the model used to generate the transcript.
            info (AsrTranscriptionInfo): Information about the transcription.
            transcription (AsrTranscription): The full transcription.
            segments (AsrSegments): Segments of the transcription.

        Returns:
            TranscriptEntity: A new instance of the TranscriptEntity class.
        """
        return TranscriptEntity(
            model=model_name,
            language=info.language,
            language_confidence=info.language_confidence,
            transcript=transcription.text,
            segments=[s.model_dump() for s in segments],
        )



import asyncio
import os
from collections.abc import AsyncGenerator
from enum import Enum
from pathlib import Path
from typing import Any, cast

import nvidia.cudnn.lib
import torch
from faster_whisper import BatchedInferencePipeline, WhisperModel
from pydantic import BaseModel, ConfigDict, Field
from ray import serve
from typing_extensions import TypedDict

from aana.core.models.asr import (
    AsrSegment,
    AsrTranscription,
    AsrTranscriptionInfo,
)
from aana.core.models.audio import Audio
from aana.core.models.base import pydantic_protected_fields
from aana.core.models.vad import VadSegment
from aana.core.models.whisper import BatchedWhisperParams, WhisperParams
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.exceptions.runtime import InferenceException

# Workaround for CUDNN issue with cTranslate2:
cudnn_path = str(Path(nvidia.cudnn.lib.__file__).parent)
os.environ["LD_LIBRARY_PATH"] = (
    cudnn_path + "/:" + os.environ.get("LD_LIBRARY_PATH", "")
)


class WhisperComputeType(str, Enum):
    """The data type used by whisper models.

    See [cTranslate2 docs on quantization](https://opennmt.net/CTranslate2/quantization.html#quantize-on-model-conversion)
    for more information.

    Available types:
        - INT8
        - INT8_FLOAT32
        - INT8_FLOAT16
        - INT8_BFLOAT16
        - INT16
        - FLOAT16
        - BFLOAT16
        - FLOAT32
    """

    INT8 = "int8"
    INT8_FLOAT32 = "int8_float32"
    INT8_FLOAT16 = "int8_float16"
    INT8_BFLOAT16 = "int8_bfloat16"
    INT16 = "int16"
    FLOAT16 = "float16"
    BFLOAT16 = "bfloat16"
    FLOAT32 = "float32"


class WhisperModelSize(str, Enum):
    """The whisper model.

    Available models:
        - TINY
        - TINY_EN
        - BASE
        - BASE_EN
        - SMALL
        - SMALL_EN
        - MEDIUM
        - MEDIUM_EN
        - LARGE_V1
        - LARGE_V2
        - LARGE_V3
        - LARGE
        - TURBO
    """

    TINY = "tiny"
    TINY_EN = "tiny.en"
    BASE = "base"
    BASE_EN = "base.en"
    SMALL = "small"
    SMALL_EN = "small.en"
    MEDIUM = "medium"
    MEDIUM_EN = "medium.en"
    LARGE_V1 = "large-v1"
    LARGE_V2 = "large-v2"
    LARGE_V3 = "large-v3"
    LARGE = "large"
    TURBO = "mobiuslabsgmbh/faster-whisper-large-v3-turbo"  # SYSTRAN/faster-whisper do not have turbo model in HF.


class WhisperConfig(BaseModel):
    """The configuration for the whisper deployment from faster-whisper.

    Attributes:
        model_size (WhisperModelSize | str): The whisper model size. Defaults to WhisperModelSize.TURBO.
        compute_type (WhisperComputeType): The compute type. Defaults to WhisperComputeType.FLOAT16.
    """

    model_size: WhisperModelSize | str = Field(
        default=WhisperModelSize.TURBO,
        description="The whisper model size. Either one of WhisperModelSize or HuggingFace model ID in Ctranslate2 format.",
    )
    compute_type: WhisperComputeType = Field(
        default=WhisperComputeType.FLOAT16, description="The compute type."
    )

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


class WhisperOutput(TypedDict):
    """The output of the whisper model.

    Attributes:
        segments (list[AsrSegment]): The ASR segments.
        transcription_info (AsrTranscriptionInfo): The ASR transcription info.
        transcription (AsrTranscription): The ASR transcription.
    """

    segments: list[AsrSegment]
    transcription_info: AsrTranscriptionInfo
    transcription: AsrTranscription


class WhisperBatchOutput(TypedDict):
    """The output of the whisper model for a batch of inputs.

    Attributes:
        segments (list[list[AsrSegment]]): The ASR segments for each audio.
        transcription_info (list[AsrTranscriptionInfo]): The ASR transcription info for each audio.
        transcription (list[AsrTranscription]): The ASR transcription for each audio.
    """

    segments: list[list[AsrSegment]]
    transcription_info: list[AsrTranscriptionInfo]
    transcription: list[AsrTranscription]


@serve.deployment(max_ongoing_requests=1)
class WhisperDeployment(BaseDeployment):
    """Deployment to serve Whisper models from faster-whisper."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        The configuration should conform to the WhisperConfig schema.
        """
        config_obj = WhisperConfig(**config)
        self.model_size = config_obj.model_size
        self.model_name = "whisper_" + self.model_size
        self.compute_type = config_obj.compute_type
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = WhisperModel(
            self.model_size, device=self.device, compute_type=self.compute_type
        )
        self.batched_model = BatchedInferencePipeline(
            model=self.model,
        )

    async def check_health(self):
        """Check the health of the deployment and clear torch cache to prevent memory leaks."""
        torch.cuda.empty_cache()
        await super().check_health()

    async def transcribe(
        self, audio: Audio, params: WhisperParams | None = None
    ) -> WhisperOutput:
        """Transcribe the audio with the whisper model.

        Args:
            audio (Audio): The audio to transcribe.
            params (WhisperParams): The parameters for the whisper model.

        Returns:
            WhisperOutput: The transcription output as a dictionary:

                - segments (List[AsrSegment]): The ASR segments.
                - transcription_info (AsrTranscriptionInfo): The ASR transcription info.
                - transcription (AsrTranscription): The ASR transcription.

        Raises:
            InferenceException: If the inference fails.
        """
        asr_segments = []
        asr_transcription_info = None
        async for output in self.transcribe_stream(audio, params):
            asr_segments.extend(output["segments"])
            asr_transcription_info = output["transcription_info"]

        transcription = "".join([seg.text for seg in asr_segments])
        asr_transcription = AsrTranscription(text=transcription)

        return WhisperOutput(
            segments=asr_segments,
            transcription_info=asr_transcription_info,
            transcription=asr_transcription,
        )

    @exception_handler
    async def transcribe_stream(
        self, audio: Audio, params: WhisperParams | None = None
    ) -> AsyncGenerator[WhisperOutput, None]:
        """Transcribe the audio with the whisper model in a streaming fashion.

        Args:
            audio (Audio): The audio to transcribe.
            params (WhisperParams): The parameters for the whisper model.

        Yields:
            WhisperOutput: The transcription output as a dictionary:

                - segments (list[AsrSegment]): The ASR segments.
                - transcription_info (AsrTranscriptionInfo): The ASR transcription info.
                - transcription (AsrTranscription): The ASR transcription.
        """
        if not params:
            params = WhisperParams()
        audio_array = audio.get_numpy()
        if not audio_array.any():
            # For silent audios/no audio tracks, return empty output with language as silence
            yield WhisperOutput(
                segments=[],
                transcription_info=AsrTranscriptionInfo(
                    language="silence", language_confidence=1.0
                ),
                transcription=AsrTranscription(text=""),
            )
        else:
            try:
                segments, info = self.model.transcribe(
                    audio_array, **params.model_dump()
                )
                asr_transcription_info = AsrTranscriptionInfo.from_whisper(info)
                for segment in segments:
                    await asyncio.sleep(0)
                    asr_segments = [AsrSegment.from_whisper(segment)]
                    asr_transcription = AsrTranscription(text=segment.text)

                    yield WhisperOutput(
                        segments=asr_segments,
                        transcription_info=asr_transcription_info,
                        transcription=asr_transcription,
                    )
            except Exception as e:
                raise InferenceException(self.model_name) from e

    async def transcribe_batch(
        self, audio_batch: list[Audio], params: WhisperParams | None = None
    ) -> WhisperBatchOutput:
        """Transcribe the batch of audios with the Whisper model.

        Args:
            audio_batch (list[Audio]): The batch of audios to transcribe.
            params (WhisperParams): The parameters for the whisper model.

        Returns:
            WhisperBatchOutput: The transcription output as a dictionary:

                - segments (list[list[AsrSegment]]): The ASR segments for each audio.
                - transcription_info (list[AsrTranscriptionInfo]): The ASR transcription info for each audio.
                - transcription (list[AsrTranscription]): The ASR transcription for each audio.

        Raises:
            InferenceException: If the inference fails.
        """
        if not params:
            params = WhisperParams()
        segments: list[list[AsrSegment]] = []
        infos: list[AsrTranscriptionInfo] = []
        transcriptions: list[AsrTranscription] = []
        for audio in audio_batch:
            output = await self.transcribe(audio, params)
            segments.append(cast(list[AsrSegment], output["segments"]))
            infos.append(cast(AsrTranscriptionInfo, output["transcription_info"]))
            transcriptions.append(cast(AsrTranscription, output["transcription"]))

        return WhisperBatchOutput(
            segments=segments, transcription_info=infos, transcription=transcriptions
        )

    @exception_handler
    async def transcribe_in_chunks(
        self,
        audio: Audio,
        vad_segments: list[VadSegment] | None = None,
        batch_size: int = 4,
        params: BatchedWhisperParams | None = None,
    ) -> AsyncGenerator[WhisperOutput, None]:
        """Transcribe a single audio by segmenting it into chunks (4x faster) in streaming mode.

        Args:
            audio (Audio): The audio to transcribe.
            vad_segments (list[VadSegment]| None): List of vad segments to guide batching of the audio data.
            batch_size (int): Maximum batch size for the batched inference.
            params (BatchedWhisperParams | None): The parameters for the batched inference with whisper model.

        Yields:
            WhisperOutput: The transcription output as a dictionary:
                segments (list[AsrSegment]): The ASR segments.
                transcription_info (AsrTranscriptionInfo): The ASR transcription info.
                transcription (AsrTranscription): The ASR transcription.

        Raises:
            InferenceException: If the inference fails.
        """
        if not params:
            params = BatchedWhisperParams()
        audio_array = audio.get_numpy()
        if vad_segments:
            vad_input = [seg.to_whisper_dict() for seg in vad_segments]
        if not audio_array.any():
            # For silent audios/no audio tracks, return empty output with language as silence
            yield WhisperOutput(
                segments=[],
                transcription_info=AsrTranscriptionInfo(
                    language="silence", language_confidence=1.0
                ),
                transcription=AsrTranscription(text=""),
            )
        else:
            try:
                segments, info = self.batched_model.transcribe(
                    audio_array,
                    vad_segments=vad_input if vad_segments else None,
                    batch_size=batch_size,
                    **params.model_dump(),
                )
                asr_transcription_info = AsrTranscriptionInfo.from_whisper(info)
                for segment in segments:
                    await asyncio.sleep(0)
                    asr_segments = [AsrSegment.from_whisper(segment)]
                    asr_transcription = AsrTranscription(text=segment.text)
                    yield WhisperOutput(
                        segments=asr_segments,
                        transcription_info=asr_transcription_info,
                        transcription=asr_transcription,
                    )
            except Exception as e:
                raise InferenceException(self.model_name) from e



import asyncio
from typing import Any

import ray
from pydantic import BaseModel, Field
from ray import serve

from aana.configs.settings import settings as aana_settings
from aana.deployments.base_deployment import BaseDeployment
from aana.storage.models.task import Status as TaskStatus
from aana.storage.repository.task import TaskRepository
from aana.storage.session import get_session
from aana.utils.core import sleep_exponential_backoff


class TaskQueueConfig(BaseModel):
    """The configuration for the task queue deployment."""

    app_name: str = Field(description="The name of the Aana app")
    retryable_exceptions: list[str] = Field(
        description="The list of exceptions that should be retried"
    )


@serve.deployment
class TaskQueueDeployment(BaseDeployment):
    """Deployment to serve the task queue.

    IMPORTANT: If you are using SQLite, make sure to run only one instance
    of this deployment to avoid database race conditions.
    """

    def __init__(self):
        """Initialize the task queue deployment."""
        super().__init__()
        loop = asyncio.get_running_loop()
        self.loop_task = loop.create_task(self.loop())
        self.loop_task.add_done_callback(
            lambda fut: fut.result() if not fut.cancelled() else None
        )
        self.running_task_ids: list[str] = []
        self.deployment_responses = {}

    def check_health(self):
        """Check the health of the deployment."""
        # if the loop is not running, the deployment is unhealthy
        if self.loop_task.done():
            raise RuntimeError(  # noqa: TRY003
                "Task queue loop is not running"
            ) from self.loop_task.exception()

    def __del__(self):
        """Clean up the deployment."""
        # Cancel the loop task to prevent tasks from being reassigned
        self.loop_task.cancel()
        # Cancel all deployment responses to stop the tasks
        # and set all non-completed tasks to NOT_FINISHED
        for task_id in self.running_task_ids:
            deployment_response = self.deployment_responses.get(task_id)
            if deployment_response:
                deployment_response.cancel()
            with get_session() as session:
                TaskRepository(session).update_status(
                    task_id=task_id,
                    status=TaskStatus.NOT_FINISHED,
                    progress=0,
                )

    async def app_health_check(self) -> bool:
        """Check the health of the app.

        The app is considered healthy if for every deployment, at least 50% of the replicas are running.
        The reason for this is that even if some replicas are not running, the app can still can process requests.
        And in the cluster setup, it is possible that some replicas on other nodes are not running or just starting up
        and it is not a reason to consider the app unhealthy.

        Returns:
            bool: True if the app is healthy, False otherwise
        """
        serve_status = serve.status()
        for app in serve_status.applications.values():
            for deployment in app.deployments.values():
                num_replicas = sum(deployment.replica_states.values())
                if num_replicas > 0:
                    health_ratio = (
                        deployment.replica_states.get("RUNNING", 0) / num_replicas
                    )
                    if health_ratio < 0.5:
                        return False
        return True

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        The configuration should conform to the TaskQueueConfig schema.
        """
        config_obj = TaskQueueConfig(**config)
        self.app_name = config_obj.app_name
        self.retryable_exceptions = config_obj.retryable_exceptions

    async def loop(self):  # noqa: C901
        """The main loop for the task queue deployment.

        The loop will check the queue and assign tasks to workers.
        """
        handle = None
        app_health_check_attempts = 0
        configuration_attempts = 0
        full_queue_attempts = 0
        no_tasks_attempts = 0
        no_app_attempts = 0

        while True:
            # Check the health of the app
            app_health = await self.app_health_check()
            if not app_health:
                # If the app is not healthy, wait and retry
                await sleep_exponential_backoff(1.0, 5.0, app_health_check_attempts)
                app_health_check_attempts += 1
                continue
            else:
                app_health_check_attempts = 0

            if not self._configured:
                # Wait for the deployment to be configured.
                await sleep_exponential_backoff(1.0, 5.0, configuration_attempts)
                configuration_attempts += 1
                continue
            else:
                configuration_attempts = 0

            with get_session() as session:
                # Remove completed tasks from the list of running tasks
                self.running_task_ids = TaskRepository(session).filter_incomplete_tasks(
                    self.running_task_ids
                )

                # Check for expired tasks
                expired_tasks = TaskRepository(session).update_expired_tasks(
                    execution_timeout=aana_settings.task_queue.execution_timeout,
                    heartbeat_timeout=aana_settings.task_queue.heartbeat_timeout,
                    max_retries=aana_settings.task_queue.max_retries,
                )
                for task in expired_tasks:
                    deployment_response = self.deployment_responses.get(task.id)
                    if deployment_response:
                        deployment_response.cancel()

            # If the queue is full, wait and retry
            if len(self.running_task_ids) >= aana_settings.task_queue.num_workers:
                await sleep_exponential_backoff(0.1, 5.0, full_queue_attempts)
                full_queue_attempts += 1
                continue
            else:
                full_queue_attempts = 0

            if not handle:
                # Try to get the app handle, if it fails, wait and retry later
                try:
                    handle = serve.get_app_handle(self.app_name)
                except ray.serve.exceptions.RayServeException as e:
                    print(f"App {self.app_name} not available yet: {e}, retrying...")
                    await sleep_exponential_backoff(1.0, 5.0, no_app_attempts)
                    no_app_attempts += 1
                    continue

            # Get new tasks from the database
            with get_session() as session:
                num_tasks_to_assign = aana_settings.task_queue.num_workers - len(
                    self.running_task_ids
                )
                tasks = TaskRepository(session).fetch_unprocessed_tasks(
                    limit=num_tasks_to_assign,
                    max_retries=aana_settings.task_queue.max_retries,
                    retryable_exceptions=self.retryable_exceptions,
                )

                # If there are no tasks, wait and retry
                if not tasks:
                    await sleep_exponential_backoff(0.1, 5.0, no_tasks_attempts)
                    no_tasks_attempts += 1
                    continue
                else:
                    no_tasks_attempts = 0

                # Start processing the tasks
                for task in tasks:
                    deployment_response = handle.execute_task.remote(task_id=task.id)
                    self.deployment_responses[task.id] = deployment_response
                    self.running_task_ids.append(str(task.id))



import contextlib
from collections.abc import AsyncGenerator

from typing_extensions import TypedDict

with contextlib.suppress(ImportError):
    pass

from aana.core.chat.chat_template import apply_chat_template
from aana.core.models.chat import ChatDialog, ChatMessage
from aana.core.models.sampling import SamplingParams
from aana.deployments.base_deployment import BaseDeployment, exception_handler


class LLMOutput(TypedDict):
    """The output of the LLM model.

    Attributes:
        text (str): the generated text
    """

    text: str


class LLMBatchOutput(TypedDict):
    """The output of the LLM model for a batch of inputs.

    Attributes:
        texts (List[str]): the list of generated texts
    """

    texts: list[str]


class ChatOutput(TypedDict):
    """The output of the chat model.

    Attributes:
        dialog (ChatDialog): the dialog with the responses from the model
    """

    message: ChatMessage


class BaseTextGenerationDeployment(BaseDeployment):
    """Base class for text generation deployments.

    To create a new text generation deployment, inherit from this class and implement at least two methods:

    `apply_config`: a method to load the model and configure it

    `generate_stream`: a method to generate text in a streaming fashion given a prompt and sampling parameters

    It defines the interface for text generation deployments and provides a default implementation for the following methods:
    `generate`, `generate_batch`, `chat`, `chat_stream`.

    You can also override these methods to implement custom inference logic.
    """

    @exception_handler
    async def generate_stream(
        self, prompt: str, sampling_params: SamplingParams | None = None
    ) -> AsyncGenerator[LLMOutput, None]:
        """Generate completion for the given prompt and stream the results.

        Args:
            prompt (str): the prompt
            sampling_params (SamplingParams | None): the sampling parameters

        Yields:
            LLMOutput: the dictionary with the key "text" and the generated text as the value
        """
        raise NotImplementedError

    @exception_handler
    async def generate(
        self, prompt: str, sampling_params: SamplingParams | None = None
    ) -> LLMOutput:
        """Generate completion for the given prompt.

        Args:
            prompt (str): the prompt
            sampling_params (SamplingParams | None): the sampling parameters

        Returns:
            LLMOutput: the dictionary with the key "text" and the generated text as the value
        """
        generated_text = ""
        async for chunk in self.generate_stream(prompt, sampling_params):
            generated_text += chunk["text"]
        return LLMOutput(text=generated_text)

    @exception_handler
    async def generate_batch(
        self, prompts: list[str], sampling_params: SamplingParams | None = None
    ) -> LLMBatchOutput:
        """Generate completion for the batch of prompts.

        Args:
            prompts (List[str]): the prompts
            sampling_params (SamplingParams | None): the sampling parameters

        Returns:
            LLMBatchOutput: the dictionary with the key "texts"
                            and the list of generated texts as the value
        """
        texts = []
        for prompt in prompts:
            text = await self.generate(prompt, sampling_params)
            texts.append(text["text"])

        return LLMBatchOutput(texts=texts)

    @exception_handler
    async def chat(
        self, dialog: ChatDialog, sampling_params: SamplingParams | None = None
    ) -> ChatOutput:
        """Chat with the model.

        Args:
            dialog (ChatDialog): the dialog
            sampling_params (SamplingParams | None): the sampling parameters

        Returns:
            ChatOutput: the dictionary with the key "message"
                        and the response message with a role "assistant"
                        and the generated text as the content
        """
        prompt = apply_chat_template(self.tokenizer, dialog, self.chat_template_name)
        response = await self.generate(prompt, sampling_params)
        response_message = ChatMessage(content=response["text"], role="assistant")
        return ChatOutput(message=response_message)

    @exception_handler
    async def chat_stream(
        self, dialog: ChatDialog, sampling_params: SamplingParams | None = None
    ) -> AsyncGenerator[LLMOutput, None]:
        """Chat with the model and stream the responses.

        Args:
            dialog (ChatDialog): the dialog
            sampling_params (SamplingParams | None): the sampling parameters

        Yields:
            LLMOutput: the dictionary with the key "text" and the generated text as the value
        """
        prompt = apply_chat_template(self.tokenizer, dialog, self.chat_template_name)
        async for chunk in self.generate_stream(prompt, sampling_params):
            yield chunk



from typing import Annotated

from pydantic import Field

__all__ = ["Caption", "CaptionsList"]

Caption = Annotated[str, Field(description="A caption.")]
"""
A caption.
"""

CaptionsList = Annotated[list[Caption], Field(description="A list of captions.")]
"""
A list of captions.
"""



from copy import copy, deepcopy
from typing import Any

import torch
from pydantic import BaseModel, ConfigDict
from ray import serve
from transformers import pipeline

from aana.core.models.base import pydantic_protected_fields
from aana.core.models.custom_config import CustomConfig
from aana.core.models.image import Image
from aana.deployments.base_deployment import BaseDeployment, exception_handler


class HfPipelineConfig(BaseModel):
    """The configuration for the Hugging Face Pipeline deployment.

    Attributes:
        model_id (str): The model ID on Hugging Face.
        task (str | None): The task name. If not provided, the task is inferred from the model ID. Defaults to None.
        model_kwargs (CustomConfig): The model keyword arguments. Defaults to {}.
        pipeline_kwargs (CustomConfig): The pipeline keyword arguments. Defaults to {}.
        generation_kwargs (CustomConfig): The generation keyword arguments. Defaults to {}.
    """

    model_id: str
    task: str | None = None
    model_kwargs: CustomConfig = {}
    pipeline_kwargs: CustomConfig = {}
    generation_kwargs: CustomConfig = {}

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


@serve.deployment
class HfPipelineDeployment(BaseDeployment):
    """Deployment to serve Hugging Face pipelines."""

    def load_pipeline(self):
        """Load the pipeline from Hugging Face."""
        self.pipeline = pipeline(
            task=self.task,
            model=self.model_id,
            model_kwargs=copy(self.model_kwargs),
            **self.pipeline_kwargs,
        )

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the pipeline from HuggingFace.

        The configuration should conform to the HfPipelineConfig schema.
        """
        config_obj = HfPipelineConfig(**config)
        self.model_id = config_obj.model_id
        self.task = config_obj.task
        self.model_kwargs = config_obj.model_kwargs
        self.pipeline_kwargs = config_obj.pipeline_kwargs
        self.generation_kwargs = config_obj.generation_kwargs

        if "device" in self.pipeline_kwargs or "device_map" in self.pipeline_kwargs:
            self.load_pipeline()  # Load pipeline with the provided device or device_map
        else:
            try:
                # If no device or device_map is provided, try to load pipeline with device_map=auto to use accelerate
                self.pipeline_kwargs["device_map"] = "auto"
                self.load_pipeline()
            except ValueError as e:
                if "does not support `device_map='auto'`" in str(e):
                    # If model doesn't support device_map=auto, use torch to figure out where to load the model
                    device = torch.device(
                        "cuda" if torch.cuda.is_available() else "cpu"
                    )
                    del self.pipeline_kwargs["device_map"]
                    self.pipeline_kwargs["device"] = device
                    self.load_pipeline()
                else:
                    raise

    @exception_handler
    async def call(self, *args, **kwargs):
        """Call the pipeline.

        Args:
            args: the input arguments
            kwargs: the input keyword arguments

        Returns:
            the output of the pipeline
        """

        def convert_images(obj):
            """Convert Aana Image objects to PIL images."""
            if isinstance(obj, Image):
                return obj.get_pil_image()
            elif isinstance(obj, list):
                return [convert_images(item) for item in obj]
            return obj

        # Apply the conversion to args
        args = [convert_images(arg) for arg in args]

        # Apply the conversion to kwargs
        kwargs = {k: convert_images(v) for k, v in kwargs.items()}

        # Update default generation kwargs with the provided ones
        _generation_kwargs = deepcopy(self.generation_kwargs)
        _generation_kwargs.update(kwargs)
        return self.pipeline(*args, **_generation_kwargs)



import asyncio
from collections.abc import Callable, Iterator
from concurrent.futures import ThreadPoolExecutor
from typing import Any

import numpy as np

__all__ = ["BatchProcessor"]


class BatchProcessor:
    """Class for parallel processing data in chunks.

    The BatchProcessor class encapsulates the logic required to take a large collection of data,
    split it into manageable batches, process these batches in parallel, and then combine the results
    into a single cohesive output.

    Batching works by iterating through the input request, which is a dictionary where each key maps
    to a list-like collection of data. The class splits each collection into sublists of length up
    to `batch_size`, ensuring that corresponding elements across the collections remain grouped
    together in their respective batches.

    Merging takes the output from each processed batch, which is also a dictionary structure, and
    combines these into a single dictionary. Lists are extended, numpy arrays are concatenated, and
    dictionaries are updated. If a new data type is encountered, an error is raised prompting the
    implementer to specify how these types should be merged.

    This class is particularly useful for batching of requests to a machine learning model.

    The thread pool for parallel processing is managed internally and is shut down automatically when
    the BatchProcessor instance is garbage collected.

    Attributes:
        process_batch (Callable): A function to process each batch.
        batch_size (int): The size of each batch to be processed.
        num_threads (int): The number of threads in the thread pool for parallel processing.
    """

    def __init__(self, process_batch: Callable, batch_size: int, num_threads: int):
        """Constructor.

        Args:
            process_batch (Callable): Function that processes each batch.
            batch_size (int): Size of the batches.
            num_threads (int): Number of threads in the pool.
        """
        self.process_batch = process_batch
        self.batch_size = batch_size
        self.pool = ThreadPoolExecutor(num_threads)

    def __del__(self):
        """Destructor that shuts down the thread pool when the instance is destroyed."""
        self.pool.shutdown()

    def batch_iterator(self, request: dict[str, Any]) -> Iterator[dict[str, list[Any]]]:
        """Converts request into an iterator of batches.

        Iterates over the input request, breaking it into smaller batches for processing.
        Each batch is a dictionary with the same keys as the input request, but the values
        are sublists containing only the elements for that batch.

        Example:
        ```python
        request = {
            'images': [img1, img2, img3, img4, img5],
            'texts': ['text1', 'text2', 'text3', 'text4', 'text5']
        }
        # Assuming a batch size of 2, this iterator would yield:
        # 1st iteration: {'images': [img1, img2], 'texts': ['text1', 'text2']}
        # 2nd iteration: {'images': [img3, img4], 'texts': ['text3', 'text4']}
        # 3rd iteration: {'images': [img5], 'texts': ['text5']}
        ```

        Args:
            request (dict[str, list[Any]]): The request data to split into batches.

        Yields:
            Iterator[dict[str, list[Any]]]: An iterator over the batched requests.
        """
        lengths = [len(value) for value in request.values()]
        if len(set(lengths)) > 1:
            raise ValueError("All inputs must have the same length")  # noqa: TRY003

        total_batches = (max(lengths) + self.batch_size - 1) // self.batch_size
        for i in range(total_batches):
            start = i * self.batch_size
            end = start + self.batch_size
            yield {key: value[start:end] for key, value in request.items()}

    async def process(self, request: dict[str, Any]) -> dict[str, Any]:
        """Process a request.

        Splits the input request into batches, processes each batch in parallel, and then merges
        the results into a single dictionary.

        Args:
            request (Dict[str, Any]): The request data to process.

        Returns:
            Dict[str, Any]: The merged results from processing all batches.
        """
        loop = asyncio.get_running_loop()
        futures = [
            loop.run_in_executor(self.pool, self.process_batch, batch)
            for batch in self.batch_iterator(request)
        ]
        outputs = await asyncio.gather(*futures)
        return self.merge_outputs(outputs)

    def merge_outputs(self, outputs: list[dict[str, Any]]) -> dict[str, Any]:
        """Merge output.

        Combine processed batch outputs into a single dictionary. It handles various data types
        by extending lists, updating dictionaries, and concatenating numpy arrays.

        Example:
        ```python
        outputs = [
            {'images': [processed_img1, processed_img2], 'labels': ['cat', 'dog']},
            {'images': [processed_img3, processed_img4], 'labels': ['bird', 'mouse']},
            {'images': [processed_img5], 'labels': ['fish']}
        ]
        # The merged result would be:
        # {
        #     'images': [processed_img1, processed_img2, processed_img3, processed_img4, processed_img5],
        #     'labels': ['cat', 'dog', 'bird', 'mouse', 'fish']
        # }
        ```

        Args:
            outputs (list[dict[str, Any]]): List of outputs from the processed batches.

        Returns:
            dict[str, Any]: The merged result.
        """
        merged_output = {}
        for output in outputs:
            for key, value in output.items():
                if key not in merged_output:
                    merged_output[key] = value
                else:
                    if isinstance(value, list):
                        merged_output[key].extend(value)
                    elif isinstance(value, dict):
                        merged_output[key].update(value)
                    elif isinstance(value, np.ndarray):
                        if key in merged_output:
                            merged_output[key] = np.concatenate(
                                (merged_output[key], value)
                            )
                        else:
                            merged_output[key] = value
                    else:
                        raise NotImplementedError(
                            "Merging of this data type is not implemented"
                        )
        return merged_output



# ruff: noqa: A002
from collections.abc import Iterable
from typing import Generic, TypeVar
from uuid import UUID

from sqlalchemy.orm import Session

from aana.core.models.media import MediaId
from aana.exceptions.db import NotFoundException
from aana.storage.models.base import BaseEntity

T = TypeVar("T", bound=BaseEntity)


# Does not yet have an update method because I'm not sure if we'll need one.
class BaseRepository(Generic[T]):
    """Base class for repositories."""

    session: Session
    table_name: str
    model_class: type[T]

    def __init__(self, session: Session, model_class: type[T]):
        """Constructor."""
        self.session = session
        self.table_name = model_class.__tablename__
        self.model_class = model_class

    def create(self, entity: T) -> T:
        """Inserts a single new entity."""
        self.session.add(entity)
        self.session.commit()
        return entity

    def create_multiple(self, entities: Iterable[T]) -> list[T]:
        """Inserts multiple entities.

        Returns:
            list[T] - entities as a list.
        """
        entities = list(entities)
        self.session.add_all(entities)
        self.session.commit()
        return entities

    def read(self, item_id: int | MediaId | UUID, check: bool = True) -> T:
        """Reads a single item by id from the database.

        Args:
            item_id (int | MediaId | UUID): id of the item to retrieve
            check (bool): whether to raise if the entity is not found (defaults to True).

        Returns:
            The corresponding entity from the database if found.

        Raises:
            NotFoundException: The id does not correspond to a record in the database.
        """
        entity: T | None = self.session.query(self.model_class).get(item_id)
        if not entity and check:
            raise NotFoundException(self.table_name, item_id)
        return entity

    def delete(self, id: int | MediaId | UUID, check: bool = True) -> T | None:
        """Deletes an entity.

        Args:
            id (int | MediaId | UUID): the id of the item to be deleted.
            check (bool): whether to raise if the entity is not found (defaults to True).

        Returns:
            The entity, if found. None, if not and `check` is False.

        Raises:
            NotFoundException if the entity is not found and `check` is True.
        """
        entity = self.read(id, check=False)
        if entity:
            self.session.delete(entity)
            self.session.flush()
            self.session.commit()
            return entity
        elif check:
            raise NotFoundException(self.table_name, id)
        else:
            return None



from typing import Any

import numpy as np
from pydantic import AliasChoices, BaseModel, ConfigDict, Field
from ray import serve
from sentence_transformers import SentenceTransformer
from typing_extensions import TypedDict

from aana.core.models.base import pydantic_protected_fields
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.exceptions.runtime import InferenceException
from aana.processors.batch import BatchProcessor


class SentenceTransformerConfig(BaseModel):
    """The configuration for the BLIP2 deployment with HuggingFace models.

    Attributes:
        model_id (str): the model ID on HuggingFace
        batch_size (int): the batch size (optional, default: 1)
        num_processing_threads (int): the number of processing threads (optional, default: 1)
    """

    model_id: str = Field(validation_alias=AliasChoices("model_id", "model"))
    batch_size: int = Field(default=1)
    num_processing_threads: int = Field(default=1)

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


class SentenceTransformerOutput(TypedDict):
    """The output of the sentence transformer model.

    Attributes:
        embedding (np.ndarray): the embedding
    """

    embedding: np.ndarray


@serve.deployment
class SentenceTransformerDeployment(BaseDeployment):
    """Deployment to serve sentence transformer models."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model from HuggingFace.

        The configuration should conform to the SentenceTransformerConfig schema.
        """
        config_obj = SentenceTransformerConfig(**config)

        # Create the batch processor to split the requests into batches
        # and process them in parallel
        self.batch_size = config_obj.batch_size
        self.num_processing_threads = config_obj.num_processing_threads
        # The actual inference is done in _generate()
        # We use lambda because BatchProcessor expects dict as input
        # and we use **kwargs to unpack the dict into named arguments for _generate()
        self.batch_processor = BatchProcessor(
            process_batch=lambda request: self._generate(**request),
            batch_size=self.batch_size,
            num_threads=self.num_processing_threads,
        )

        self.model_id = config_obj.model_id

        self.model = SentenceTransformer(self.model_id)

    @exception_handler
    async def embed_batch(self, **kwargs) -> np.ndarray:
        """Embed the given sentences.

        Args:
            sentences (List[str]): the sentences
            **kwargs (dict[str, Any]): keywordarguments to pass to the
                batch processor.

        Returns:
            np.ndarray: the embeddings for the sentences

        Raises:
            InferenceException: if the inference fails
        """
        # Call the batch processor to process the requests
        # The actual inference is done in _generate()
        return await self.batch_processor.process(kwargs)

    def _generate(self, sentences: list[str]) -> SentenceTransformerOutput:
        """Generate the embeddings for the given sentences.

        This method is called by the batch processor.

        Args:
            sentences (list[str]): the sentences

        Returns:
            SentenceTransformerOutput: the dictionary with one key "embedding"

        Raises:
            InferenceException: if the inference fails
        """
        try:
            embeddings = self.model.encode(sentences)
            return SentenceTransformerOutput(embedding=embeddings)
        except Exception as e:
            raise InferenceException(self.model_id) from e



from typing import TypeVar

from sqlalchemy.orm import Session

from aana.core.models.media import MediaId
from aana.exceptions.db import MediaIdAlreadyExistsException, NotFoundException
from aana.storage.models.media import MediaEntity
from aana.storage.repository.base import BaseRepository

M = TypeVar("M", bound=MediaEntity)


class MediaRepository(BaseRepository[M]):
    """Repository for media files."""

    def __init__(self, session: Session, model_class: type[M] = MediaEntity):
        """Constructor."""
        super().__init__(session, model_class)

    def check_media_exists(self, media_id: MediaId) -> bool:
        """Checks if a media file exists in the database.

        Args:
            media_id (MediaId): The media ID.

        Returns:
            bool: True if the media exists, False otherwise.
        """
        try:
            self.read(media_id)
        except NotFoundException:
            return False

        return True

    def create(self, entity: MediaEntity) -> MediaEntity:
        """Inserts a single new entity.

        Args:
            entity (MediaEntity): The entity to insert.

        Returns:
            MediaEntity: The inserted entity.
        """
        # TODO: throw MediaIdAlreadyExistsException without checking if media exists.
        # The following code is a better way to check if media already exists because it does only one query and
        # prevents race condition where two processes try to create the same media.
        # But it has an issue with the exception handling.
        # Unique constraint violation raises IntegrityError, but IntegrityError much more broader than
        # Unique constraint violation. So we cannot catch IntegrityError and raise MediaIdAlreadyExistsException,
        # we need to check the exception if it is Unique constraint violation or not and if it's on the media_id column.
        # Also different DBMS have different exception messages.
        #
        # try:
        #     return super().create(entity)
        # except IntegrityError as e:
        #     self.session.rollback()
        #     raise MediaIdAlreadyExistsException(self.table_name, entity.id) from e

        if self.check_media_exists(entity.id):
            raise MediaIdAlreadyExistsException(self.table_name, entity.id)

        return super().create(entity)



from ray import serve

from aana.utils.core import sleep_exponential_backoff
from aana.utils.typing import is_async_generator


class AanaDeploymentHandle:
    """A handle to interact with a deployed Aana deployment.

    Use create method to create a deployment handle.

    ```python
    deployment_handle = await AanaDeploymentHandle.create("deployment_name")
    ```

    Attributes:
        handle (ray.serve.handle.DeploymentHandle): Ray Serve deployment handle.
        deployment_name (str): The name of the deployment.
    """

    def __init__(
        self,
        deployment_name: str,
        num_retries: int = 3,
        retry_exceptions: bool | list[Exception] = False,
        retry_delay: float = 0.2,
        retry_max_delay: float = 2.0,
    ):
        """A handle to interact with a deployed Aana deployment.

        Args:
            deployment_name (str): The name of the deployment.
            num_retries (int): The maximum number of retries for the method.
            retry_exceptions (bool | list[Exception]): Whether to retry on application-level errors or a list of exceptions to retry on.
            retry_delay (float): The initial delay between retries.
            retry_max_delay (float): The maximum delay between retries.
        """
        self.handle = serve.get_app_handle(deployment_name)
        self.deployment_name = deployment_name
        self.__methods = None
        self.num_retries = num_retries
        self.retry_exceptions = retry_exceptions
        self.retry_delay = retry_delay
        self.retry_max_delay = retry_max_delay

    def __create_async_method(self, name: str):  # noqa: C901
        """Create an method to interact with the deployment.

        Args:
            name (str): The name of the method.
        """
        method_info = self.__methods[name]
        annotations = method_info.get("annotations", {})
        return_type = annotations.get("return", None)

        if is_async_generator(return_type):

            async def method(*args, **kwargs):
                retries = 0
                while retries <= self.num_retries:
                    try:
                        async for item in self.handle.options(
                            method_name=name, stream=True
                        ).remote(*args, **kwargs):
                            yield item
                        break
                    except Exception as e:
                        is_retryable = self.retry_exceptions is True or (
                            isinstance(self.retry_exceptions, list)
                            and isinstance(
                                e.cause.__class__, tuple(self.retry_exceptions)
                            )
                        )
                        if not is_retryable or retries >= self.num_retries:
                            raise
                        await sleep_exponential_backoff(
                            initial_delay=self.retry_delay,
                            max_delay=self.retry_max_delay,
                            attempts=retries,
                        )
                        retries += 1

        else:

            async def method(*args, **kwargs):
                retries = 0
                while retries <= self.num_retries:
                    try:
                        return await self.handle.options(method_name=name).remote(
                            *args, **kwargs
                        )
                    except Exception as e:  # noqa: PERF203
                        is_retryable = self.retry_exceptions is True or (
                            isinstance(self.retry_exceptions, list)
                            and isinstance(
                                e.cause.__class__, tuple(self.retry_exceptions)
                            )
                        )
                        if not is_retryable or retries >= self.num_retries:
                            raise
                        await sleep_exponential_backoff(
                            initial_delay=self.retry_delay,
                            max_delay=self.retry_max_delay,
                            attempts=retries,
                        )
                        retries += 1

        if "annotations" in self.__methods[name]:
            method.__annotations__ = self.__methods[name]["annotations"]
        if "doc" in self.__methods[name]:
            method.__doc__ = self.__methods[name]["doc"]
        return method

    async def __load_methods(self):
        """Load the methods available in the deployment."""
        self.__methods = await self.handle.get_methods.remote()
        for name in self.__methods:
            setattr(self, name, self.__create_async_method(name))

    @classmethod
    async def create(
        cls,
        deployment_name: str,
        num_retries: int = 3,
        retry_exceptions: bool | list[Exception] = False,
        retry_delay: float = 0.2,
        retry_max_delay: float = 2.0,
    ):
        """Create a deployment handle.

        Args:
            deployment_name (str): The name of the deployment to interact with.
            num_retries (int): The maximum number of retries for the method.
            retry_exceptions (bool | list[Exception]): Whether to retry on application-level errors or a list of exceptions to retry on.
            retry_delay (float): The initial delay between retries.
            retry_max_delay (float): The maximum delay between retries.
        """
        handle = cls(
            deployment_name=deployment_name,
            num_retries=num_retries,
            retry_exceptions=retry_exceptions,
            retry_delay=retry_delay,
            retry_max_delay=retry_max_delay,
        )
        await handle.__load_methods()
        return handle



from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field

from aana.core.models.base import pydantic_protected_fields

__all__ = [
    "Role",
    "Prompt",
    "Question",
    "ChatMessage",
    "ChatDialog",
    "ChatCompletionRequest",
    "ChatCompletionChoice",
    "ChatCompletion",
]

Role = Literal["system", "user", "assistant"]
"""
The role of a participant in a conversation.

- "system": Used for instructions or context provided to the model.
- "user": Represents messages from the user.
- "assistant": Represents LLM responses.
"""

Prompt = Annotated[str, Field(description="The prompt for the LLM.")]
"""
The prompt for the LLM.
"""

Question = Annotated[str, Field(description="The question.")]
"""
The question.
"""


class ChatMessage(BaseModel):
    """A chat message.

    Attributes:
        content (str): the text of the message
        role (Role): the role of the message
    """

    content: str
    role: Role
    model_config = ConfigDict(
        json_schema_extra={
            "description": "A chat message.",
            "examples": [
                {
                    "role": "system",
                    "content": "You are a helpful assistant.",
                },
                {
                    "role": "user",
                    "content": "Hello, how are you?",
                },
                {
                    "role": "assistant",
                    "content": "I am doing well, thank you.",
                },
            ],
        }
    )


class ChatDialog(BaseModel):
    """A chat dialog.

    Attributes:
        messages (list[ChatMessage]): the messages in the dialog
    """

    messages: list[ChatMessage]
    model_config = ConfigDict(
        json_schema_extra={
            "description": "A chat dialog.",
            "examples": [
                {
                    "messages": [
                        {
                            "role": "system",
                            "content": "You are a helpful assistant.",
                        },
                        {
                            "role": "user",
                            "content": "Hello, how are you?",
                        },
                        {
                            "role": "assistant",
                            "content": "I am doing well, thank you.",
                        },
                    ]
                }
            ],
        }
    )

    @classmethod
    def from_list(cls, messages: list[dict[str, str]]) -> "ChatDialog":
        """Create a ChatDialog from a list of dictionaries.

        Args:
            messages (list[dict[str, str]]): the list of messages

        Returns:
            ChatDialog: the chat dialog
        """
        return ChatDialog(messages=[ChatMessage(**message) for message in messages])

    @classmethod
    def from_prompt(cls, prompt: str) -> "ChatDialog":
        """Create a ChatDialog from a prompt.

        Args:
            prompt (str): the prompt

        Returns:
            ChatDialog: the chat dialog
        """
        return ChatDialog(messages=[ChatMessage(content=prompt, role="user")])


class ChatCompletionRequest(BaseModel):
    """A chat completion request for OpenAI compatible API.

    Attributes:
        model (str): the model name (name of the LLM deployment)
        messages (list[ChatMessage]): a list of messages comprising the conversation so far
        temperature (float): float that controls the randomness of the sampling
        top_p (float): float that controls the cumulative probability of the top tokens to consider
        max_tokens (int): the maximum number of tokens to generate
        repetition_penalty (float): float that penalizes new tokens based on whether they appear in the prompt and the generated text so far
        stream (bool): if set, partial message deltas will be sent
    """

    model: str = Field(..., description="The model name (name of the LLM deployment).")
    messages: list[ChatMessage] = Field(
        ..., description="A list of messages comprising the conversation so far."
    )
    temperature: float = Field(
        default=1.0,
        ge=0.0,
        description=(
            "Float that controls the randomness of the sampling. "
            "Lower values make the model more deterministic, "
            "while higher values make the model more random. "
            "Zero means greedy sampling."
        ),
    )
    top_p: float = Field(
        default=1.0,
        gt=0.0,
        le=1.0,
        description=(
            "Float that controls the cumulative probability of the top tokens to consider. "
            "Must be in (0, 1]. Set to 1 to consider all tokens."
        ),
    )
    max_tokens: int | None = Field(
        default=None, ge=1, description="The maximum number of tokens to generate."
    )
    repetition_penalty: float = Field(
        default=1.0,
        description=(
            "Float that penalizes new tokens based on whether they appear in the "
            "prompt and the generated text so far. Values > 1 encourage the model "
            "to use new tokens, while values < 1 encourage the model to repeat tokens. "
            "Default is 1.0 (no penalty)."
        ),
    )

    stream: bool | None = Field(
        default=False,
        description=(
            "If set, partial message deltas will be sent, like in ChatGPT. "
            "Tokens will be sent as data-only server-sent events as they become available, "
            "with the stream terminated by a data: [DONE] message."
        ),
    )

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


class ChatCompletionChoice(BaseModel):
    """A chat completion choice for OpenAI compatible API.

    Attributes:
        index (int): the index of the choice in the list of choices
        message (ChatMessage): a chat completion message generated by the model
    """

    index: int = Field(
        ..., description="The index of the choice in the list of choices."
    )
    message: ChatMessage = Field(
        ..., description="A chat completion message generated by the model."
    )


class ChatCompletion(BaseModel):
    """A chat completion for OpenAI compatible API.

    Attributes:
        id (str): a unique identifier for the chat completion
        model (str): the model used for the chat completion
        created (int): the Unix timestamp (in seconds) of when the chat completion was created
        choices (list[ChatCompletionChoice]): a list of chat completion choices
        object (Literal["chat.completion"]): the object type, which is always `chat.completion`
    """

    id: str = Field(..., description="A unique identifier for the chat completion.")
    model: str = Field(..., description="The model used for the chat completion.")
    created: int = Field(
        ...,
        description="The Unix timestamp (in seconds) of when the chat completion was created.",
    )
    choices: list[ChatCompletionChoice] = Field(
        ...,
        description="A list of chat completion choices.",
    )
    object: Literal["chat.completion"] = Field(
        "chat.completion",
        description="The object type, which is always `chat.completion`.",
    )

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))



import typing
from collections.abc import AsyncGenerator


def is_typed_dict(argument: type) -> bool:
    """Checks if a argument is a TypedDict.

    Arguments:
        argument (type): the type to check

    Returns:
        bool: True if the argument type is a TypedDict anf False if it is not.
    """
    return bool(
        argument and getattr(argument, "__orig_bases__", None) == (typing.TypedDict,)
    )


def is_async_generator(argument: type) -> bool:
    """Checks if a argument is an AsyncGenerator.

    Arguments:
        argument (type): the type to check

    Returns:
        bool: True if the argument type is an AsyncGenerator and False if it is not.
    """
    return hasattr(argument, "__origin__") and issubclass(
        argument.__origin__, AsyncGenerator
    )
